from lantz.driver import Driver
from lantz import Feat, Action
import serial
import numpy as np
import struct
import time
import enum

class MotDLpro(Driver):

    tmcl_struct = struct.Struct('>BBBBiB')

    _STATUSCODES = {
        100 : "Succesfully executed, no error",
        101 : "Command loaded into TMCL program EEPROM",
          1 : "Wrong Checksum",
          2 : "Invalid command",
          3 : "Wrong type",
          4 : "Invalid value",
          5 : "Configuration EEPROM locked",
          6 : "Command not available",
    }

    TMCL_OK_STATUS = {
        100, 101
    }

    _STAT_OK = 100

    class CMDS(enum.Enum):
        ROR = 1
        ROL = 2
        MST = 3
        MVP = 4
        SAP = 5
        GAP = 6
        STAP = 7
        RSAP = 8
        SGP = 9
        GGP = 10
        STGP = 11
        RSGP = 12
        RFS = 13
        SIO = 14
        GIO = 15
        CALC = 19
        COMP = 20
        JC = 21
        JA = 22
        CSUB = 23
        RSUB = 24
        EI = 25
        DI = 26
        WAIT = 27
        STOP = 28
        SCO = 30
        GCO = 31
        CCO = 32
        CALCX = 33
        AAP = 34
        AGP = 35
        VECT = 37
        RETI = 38
        ACO = 39

    _COMMAND_NUMBERS = {
        1 : "ROR",    2 : "ROL",    3 : "MST",
        4 : "MVP",    5 : "SAP",    6 : "GAP",
        7 : "STAP",   8 : "RSAP",   9 : "SGP",
        10 : "GGP",   11 : "STGP",  12 : "RSGP",
        13 : "RFS",   14 : "SIO",   15 : "GIO",
        19 : "CALC",  20 : "COMP",  21 : "JC",
        22 : "JA",    23 : "CSUB",  24 : "RSUB",
        25 : "EI",    26 : "DI",    27 : "WAIT",
        28 : "STOP",  30 : "SCO",   31 : "GCO",
        32 : "CCO",   33 : "CALCX", 34 : "AAP",
        35 : "AGP",   37 : "VECT",  38 : "RETI",
        39 : "ACO",
    }

    _NUMBER_COMMANDS = dict((v, k) for k, v in _COMMAND_NUMBERS.items())

    _INTERRUPT_VECTORS = {
        0 : "Timer 0",
        1 : "Timer 1",
        2 : "Timer 2",
        3 : "Target position reached",
       15 : "stallGuard",
       21 : "Deviation",
       27 : "Left stop switch",
       28 : "Right stop switch",
       39 : "Input change 0",
       40 : "Input change 1",
      255 : "Global interrupts",
    }

    _CMD_MVP_TYPES = {
        'ABS' : 0,
        'REL' : 1,
        'COORDS' : 2,
    }

    _CMD_RFS_TYPES = {
        'START' : 0,
        'STOP' : 1,
        'STATUS' : 2,
    }

    TR_24s = [(-2**23+1, 2**23)]
    TR_32u = [(0, 2**32)]
    TR_32s = [(-2**31, 2**31)]
    TR_16u = [(0, 2**16)]
    TR_12s = [(-2**11+1, 2**11)]
    TR_12u = [(0, 2**12)]
    TR_11u = [(0, 2**11)]
    TR_10u = [(0, 2**10)]
    TR_8u = [(0, 2**8)]
    TR_7s = [(-2**6, 2**6)]
    TR_5u = [(0, 2**5)]
    TR_1u = [(0, 2**1)]
    TR_m3 = [(0, 3)]
    TR_m4 = [(0, 4)]
    TR_m9 = [(0, 9)]
    TR_m12 = [(0, 14)]
    TR_m14 = [(0, 14)]
    TR_m16 = [(0, 16)]

    TR_xCHP0 = [(-3, 13)]
    TR_xCHP1 = [(0, 1), (2, 16)]
    TR_xSE0 = [(1, 4)]
    TR_xRFS0 = [(1, 9)]
    TR_xRFS1 = [(0, 8388307)]
    TR_xPWR0 = [(1, 2**16)]
    TR_xRND0 = [(0, 2**31)]

    T_R = 4
    T_W = 2
    T_E = 1
    T_RW = T_R + T_W
    T_RWE = T_RW + T_E

    _AXIS_PARAMETER = {
         0 : ("target position", TR_24s, T_RW),
         1 : ("actual position", TR_24s, T_RW),
         2 : ("target speed", TR_12s, T_RW),
         3 : ("actual speed", TR_12s, T_RW),
         4 : ("max positioning speed", TR_11u, T_RWE),
         5 : ("max acceleration", TR_11u, T_RWE),
         6 : ("abs max current", TR_8u, T_RWE),
         7 : ("standby current", TR_8u, T_RWE),
         8 : ("target pos reached", TR_1u, T_R),
         9 : ("ref switch status", TR_1u, T_R),
        10 : ("right limit switch status", TR_1u, T_R),
        11 : ("left limit switch status", TR_1u, T_R),
        12 : ("right limit switch disable", TR_1u, T_RWE),
        13 : ("left limit switch disable", TR_1u, T_RWE),
       130 : ("minimum speed", TR_11u, T_RWE),
       135 : ("actual acceleration", TR_11u, T_R),
       138 : ("ramp mode", TR_m3, T_RWE),
       140 : ("microstep resolution", TR_m9, T_RWE),
       141 : ("ref switch tolerance", TR_12u, T_RW),
       149 : ("soft stop flag", TR_1u, T_RWE),
       153 : ("ramp divisor", TR_m14, T_RWE),
       154 : ("pulse divisor", TR_m14, T_RWE),
       160 : ("step interpolation enable", TR_1u, T_RW),
       161 : ("double step enable", TR_1u, T_RW),
       162 : ("chopper blank time", TR_m4, T_RW),
       163 : ("chopper mode", TR_1u, T_RW),
       164 : ("chopper hysteresis dec", TR_m4, T_RW),
       165 : ("chopper hysteresis end", TR_xCHP0, T_RW),
       166 : ("chopper hysteresis start", TR_m9, T_RW),
       167 : ("chopper off time", TR_xCHP1, T_RW),
       168 : ("smartEnergy min current", TR_1u, T_RW),
       169 : ("smartEnergy current downstep", TR_m4, T_RW),
       170 : ("smartEnergy hysteresis", TR_m16, T_RW),
       171 : ("smartEnergy current upstep", TR_xSE0, T_RW),
       172 : ("smartEnergy hysteresis start", TR_m16, T_RW),
       173 : ("stallGuard2 filter enable", TR_1u, T_RW),
       174 : ("stallGuard2 threshold", TR_7s, T_RW),
       175 : ("slope control high side", TR_m4, T_RW),
       176 : ("slope control low side", TR_m4, T_RW),
       177 : ("short protection disable", TR_1u, T_RW),
       178 : ("short detection timer", TR_m4, T_RW),
       179 : ("Vsense", TR_1u, T_RW),
       180 : ("smartEnergy actual current", TR_5u, T_RW),
       181 : ("stop on stall", TR_11u, T_RW),
       182 : ("smartEnergy threshold speed", TR_11u, T_RW),
       183 : ("smartEnergy slow run current", TR_8u, T_RW),
       193 : ("ref. search mode", TR_xRFS0, T_RWE),
       194 : ("ref. search speed", TR_11u, T_RWE),
       195 : ("ref. switch speed", TR_11u, T_RWE),
       196 : ("distance end switches", TR_xRFS1, T_R),
       204 : ("freewheeling", TR_16u, T_RWE),
       206 : ("actual load value", TR_10u, T_R),
       208 : ("TMC262 errorflags", TR_8u, T_R),
       209 : ("encoder pos", TR_24s, T_RW),
       210 : ("encoder prescaler", TR_16u, T_RWE), # that one isnt really correct
       212 : ("encoder max deviation", TR_16u, T_RWE),
       214 : ("power down delay", TR_xPWR0, T_RWE),
    }

    _SINGLE_AXIS_PARAMETERS = [140] + list(range(160, 184))

    _GLOBAL_PARAMETER = {
        (0, 64) : ("EEPROM magic", TR_8u, T_RWE),
        (0, 65) : ("RS485 baud rate", TR_m12, T_RWE),
        (0, 66) : ("serial address", TR_8u, T_RWE),
        (0, 73) : ("EEPROM lock flag", TR_1u, T_RWE),
        (0, 75) : ("telegram pause time", TR_8u, T_RWE),
        (0, 76) : ("serial host adress", TR_8u, T_RWE),
        (0, 77) : ("auto start mode", TR_1u, T_RWE),
        (0, 81) : ("TMCL code protect", TR_m4, T_RWE),
        # (0, 84) : ("coordinate storage", TR_1u, T_RWE), # wrong type?
        (0, 128) : ("TMCL application status", TR_m3, T_R),
        (0, 129) : ("download mode", TR_1u, T_R),
        (0, 130) : ("TMCL program counter", TR_32u, T_R),
        (0, 132) : ("tick timer", TR_32u, T_RW),
        # (0, 133) : ("random number", TR_xRND0, T_R), # wrong type?
        (3, 0) : ("Timer0 period", TR_32u, T_RWE),
        (3, 1) : ("Timer1 period", TR_32u, T_RWE),
        (3, 2) : ("Timer2 period", TR_32u, T_RWE),
        (3, 39) : ("Input0 edge type", TR_m4, T_RWE),
        (3, 40) : ("Input0 edge type", TR_m4, T_RWE),
    }

    def __init__(self, address, target):
        self.address = address
        self.target = target
        return

    def initialize(self):
        self._serial = serial.Serial(
            port=self.address,
            baudrate=9600,
            bytesize=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            timeout=0.1,
        )

        self.init_motor_parameters()
        self.get_calibration_data()
        return

    def set_sap_for_all_motors(self, sap_type, sap_value):
        for mot in range(3):
            self.send_instruction(5, sap_type, mot, sap_value)
        return

    def init_motor_parameters(self):
        self.set_sap_for_all_motors(4, 1000)    # set speed to 1000
        self.set_sap_for_all_motors(5, 100)     # set max accel to 100
        self.set_sap_for_all_motors(6, 42)      # set maxcurrentl to 250 - this is in units of 255 = 100%; I think the max current is 1.5 A so this is ~250 mA.
        self.set_sap_for_all_motors(7, 0)       # set standby current to 0
        self.set_sap_for_all_motors(12, 1)      # disable right limit switch
        self.set_sap_for_all_motors(13, 0)      # enable left limit switch
        self.set_sap_for_all_motors(140, 4)     # set microstep resolution to 16 microsteps (parameter value = 4)

        # vvv this was commented
        # self.set_sap_for_all_motors(143, 1)     # set rest current to approximately 12%

        self.set_sap_for_all_motors(153, 7)     # set ramp to 7
        self.set_sap_for_all_motors(154, 3)     # set pulse to 3
        self.set_sap_for_all_motors(194, 1000)  # set reference speed to 1000
        self.set_sap_for_all_motors(203, 100)   # set the mixed decay threshold to 100

        # vvv this had a value of 30
        self.set_sap_for_all_motors(204, 10)    # set FreeWheelTime to 10

        for idx in range(6):
            self.send_instruction(14, typ=idx, mot=0, val=0)
        return

    # def get_calibration_data(self):
    #     # hardcode for now - need to change when getting new laser
    #     self.wavelength_limits = (1064.86, 1144.51)
    #     self.p_coeffs = (-1.15278e6, 64.5135, 0.962788)
    #     self.backlash_coeff = 5035
    #     return

    def toptica_integer_to_double(self, ints):
        # this is how DC got the hardcoded values in the commented get_calibration_data
        return struct.unpack('d', struct.pack('ii', *reversed(ints)))[0]

    def get_calibration_data(self):
        params = self.get_global_parameters()
        min_wl = self.toptica_integer_to_double(params[22:24])
        max_wl = self.toptica_integer_to_double(params[24:26])
        p2 = self.toptica_integer_to_double(params[26:28])
        p1 = self.toptica_integer_to_double(params[28:30])
        p0 = self.toptica_integer_to_double(params[30:32])
        backlash = params[32]
        self.wavelength_limits = (min_wl, max_wl)
        self.p_coeffs = p2, p1, p0
        self.backlash_coeff = backlash
        self.step_margin = 1000
        self.step_limits = self.wavelength_to_step(min_wl), self.wavelength_to_step(max_wl)
        return

    def checksum(self, buffer):
        chk = np.sum(buffer, dtype=np.uint8)
        return chk

    def send_instruction(self, n, typ=0, mot=0, val=0):
        """
        Sends one instruction, and return the reply.
        n (0<=int<=255): instruction ID
        typ (0<=int<=255): instruction type
        mot (0<=int<=255): motor/bank number
        val (0<=int<2**32): value to send
        return (0<=int<2**32): value of the reply (if status is good)
        raises:
            IOError: if problem with sending/receiving data over the serial port
            TMCLError: if status if bad

        IMPORTANT: TO BE COMPATIBLE WITH TOPTICA'S CODE, IF YOU LOOK IN THE
        GENERATE/SEND COMMAND VI, FOR SOME REASON THE AUTHOR TAKES THE INTEGER
        2 AND SUBTRACTS THE MOTOR NUMBER FROM IT. SO MOTOR 0 IN THE TOPTICA
        VI IS ACTUALLY MOTOR 2 IN THE TMCL CODE.
        """

        mot = 2 - mot
        data = [
            self.target,
            n,
            typ,
            mot,
            val,
            0,
        ]
        msg = np.frombuffer(self.tmcl_struct.pack(*data), dtype=np.uint8).tolist()
        msg[-1] = self.checksum(msg)
        self._serial.write(msg)
        self._serial.flush()
        while True:
            ret = self._serial.read(9)
            if len(ret) < 9:
                # could not read 9 bytes
                pass
            *retdata, chk = self.tmcl_struct.unpack(ret)
            if self.checksum(np.frombuffer(ret, dtype=np.uint8)[:-1]) == chk:
                ra, rt, status, rn, rval = retdata
                if self.target and self.target != rt: # target = 0 means any device
                    pass
                if rn != n:
                    continue
                if status not in self.TMCL_OK_STATUS:
                    raise TMCLError(status, rval)
            else:
                raise TMCLError
            return rval

    def wavelength_to_step(self, wl):
        if not self.wavelength_limits[0] <= wl <= self.wavelength_limits[1]:
            raise ValueError('wavelength {} out of operation range ({}, {})'.format(wl, *self.wavelength_limits))
        p0, p1, p2 = self.p_coeffs
        step = p0 + p1 * wl + p2 * wl * wl
        return int(step)

    def step_to_wavelength(self, step):
        p0, p1, p2 = self.p_coeffs
        wl = (-p1 + np.sqrt(p1 * p1 - 4 * p2 * (p0 - step))) / (2 * p2)
        if not self.wavelength_limits[0] <= wl <= self.wavelength_limits[1]:
            raise ValueError('wavelength {} out of operation range ({}, {})'.format(wl, *self.wavelength_limits))
        return wl

    @Feat(limits=(0, 300000))
    def position(self):
        return int(self.send_instruction(6, typ=1, mot=0, val=0))

    @position.setter
    def position(self, step):
        step = int(step)
        ret = self.send_instruction(4, typ=0, mot=0, val=step)
        time.sleep(0.25)
        now = time.time()
        while time.time() < (now + 10.0):
            if step == self.position:
                break
            else:
                time.sleep(0.1)
        else:
            raise MotDLproError('timeout, needed to get to {}, stopped at {}'.format(step, self.position))
        return

    @Feat()
    def wavelength(self):
        return self.step_to_wavelength(self.position)

    @wavelength.setter
    def wavelength(self, wl):
        target_position = self.wavelength_to_step(wl)
        current_position = self.position
        if current_position > target_position:
            # if we make a movement to the left, set the desired step
            # to beyond by the "backlash" calibration parameter
            target_position -= self.backlash_coeff
        self.position = target_position

    @Action()
    def precision_move(self, target_position, offset=10000, from_high=False):
        current_position = self.position
        offset = -offset if from_high else offset
        lower, upper = self.step_limits
        args = [
            current_position + offset,
            lower - self.step_margin,
            upper + self.step_margin,
        ]
        offset_position = np.clip(*args)
        status = True
        try:
            self.position = offset_position
            time.sleep(0.25)
            self.position = target_position
        except MotDLproError:
            status = False
        return status

    # @Action()
    # def precision_move(self, step, direction=1):
    #     if direction == 1:
    #         if step > 10000:
    #             self.position = step - 10000
    #         else:
    #             self.position = 0
    #     else:
    #         if step < 172000:
    #             self.position = step + 10000
    #         else:
    #             self.position = 172000 + 10000
    #     self.position = step
    #     return

    @Action()
    def reference_search(self):
        self.send_instruction(13, typ=0, mot=0, val=0)

        now = time.time()
        while time.time() < (now + 45.0):
            ret = self.send_instruction(13, typ=2, mot=0, val=0)
            if not ret:
                break
        else:
            print('timeout')
            self.send_instruction(13, typ=2, mot=0, val=0)
        return

    @Action()
    def calibrate(self, steps=1000):
        lower, upper = self.wavelength_limits
        _positions = np.linspace(lower, upper, steps)
        frequencies = list()
        wavelengths = list()

        for from_high in [False, True]:
            if from_high:
                positions = _positions[::-1]
            for position in positions:
                self.precision_move(position, from_high=from_high)
                frequencies.append(wavemeter.frequency)
                wavelengths.append(wavemeter.wavelength)

        frequencies = np.array(frequencies)
        wavelengths = np.array(wavelengths)

        n = len(positions)
        # we will create the calibration fit from the lower-to-upper motor direction
        p = np.polyfit(wavelength[:n], _positions, deg=2)

        def motor_pos(wl):
            return p[0] * np.square(wl) + p[1] * wl + p[2]

        fit_positions = motor_pos(wavelengths[n:])
        backlash = np.mean(fit_positions - _positions[::-1])

        wl_max = np.max(wavelengths[:n])
        wl_min = np.min(wavelengths[n:])

        return wl_min, wl_max, p[2], p[1], p[0], backlash

    def restore_global_parameters(self):
        parameters = list()
        for idx in range(34):
            offset = 20 + idx
            self.send_instruction(12, typ=offset, mot=0, val=0)
            ret = self.send_instruction(10, typ=offset, mot=0, val=0)
            parameters.append(ret)
        return parameters

    def get_global_parameters(self):
        parameters = list()
        for idx in range(34):
            offset = 20 + idx
            ret = self.send_instruction(10, typ=offset, mot=0, val=0)
            parameters.append(ret)
        return parameters


class TMCLError(Exception):
    pass

class MotDLproError(Exception):
    pass

def test():
    mot = MotDLpro('COM7', 1)
    mot.initialize()
    mot.restore_global_parameters()
    print(mot.wavelength)
    mot.wavelength += 2
    print(mot.wavelength)
    mot.wavelength -= 2
    print(mot.wavelength)

if __name__ == '__main__':
    test()



from lantz import Driver, DictFeat, Action
import requests
import re
from urllib.parse import urlparse, ParseResult

class Relay(Driver):

    def __init__(self, host='192.168.1.4', endpoint='30000'):
        super().__init__()

        p = urlparse(host, 'http')
        netloc = p.netloc or p.path
        path = p.path if p.netloc else ''
        p = ParseResult('http', netloc, path, *p[3:])
        self.host = p.geturl()
        self.endpoint = str(endpoint)
        self.status_page = requests.compat.urljoin(self.host, '{}/42'.format(self.endpoint))
        self.all_on_page = requests.compat.urljoin(self.host, '{}/45'.format(self.endpoint))
        self.all_off_page = requests.compat.urljoin(self.host, '{}/46'.format(self.endpoint))
        return

    def parse_state(self, resp_text):
        relay_words = [m.start() for m in re.finditer('Relay', resp_text)][2:]
        relay_status = [resp_text[s+33:s+36] for s in relay_words]
        relay_status = [False if v == 'OFF' else True for v in relay_status]
        return relay_status

    def get_all_states(self):
        resp1 = requests.get(self.status_page)
        resp2 = requests.get(self.status_page)
        if 'Relay-01' not in resp1:
            resp1, resp2 = resp2, resp1
        states = self.parse_state(resp1.text) + self.parse_state(resp2.text)
        return states

    @DictFeat(keys=range(1, 9), values={True, False})
    def state(self, key):
        states = self.get_all_states()
        state = states[key - 1]
        return state

    @state.setter
    def state(self, key, value):
        offset = 1 if value else 0
        page = requests.compat.urljoin(self.host, '{}/{:02d}'.format(self.endpoint, 2 * key - 2 + offset))
        requests.get(page)
        return

    @Action()
    def all_on(self):
        resp = requests.get(self.all_on_page)
        return

    @Action()
    def all_off(self):
        resp = requests.get(self.all_off_page)
        return



import numpy as np
from time import sleep


import sys
import os

try:
    import clr
    # Import DLLs for running spectrometer via LightField
    lf_root = os.environ['LIGHTFIELD_ROOT']
    automation_path = lf_root + '\PrincetonInstruments.LightField.AutomationV4.dll'
    addin_path = lf_root + '\AddInViews\PrincetonInstruments.LightFieldViewV4.dll'
    support_path = lf_root + '\PrincetonInstruments.LightFieldAddInSupportServices.dll'

    addin_class = clr.AddReference(addin_path);
    automation_class = clr.AddReference(automation_path);
    support_class = clr.AddReference(support_path);

    import PrincetonInstruments.LightField as lf

    # Import some system functions for interfacing with LightField code
    clr.AddReference("System.Collections")
    clr.AddReference("System.IO")
    from System.Collections.Generic import List
    from System import String
    from System.IO import FileAccess
except:
    pass

# Lantz imports
from lantz import Driver, Feat, DictFeat, Action



class LightFieldM:
    """
    Helper class for interfacing between the lantz driver and LightField
    software.
    """

    def __init__(self, visible):

        self._addinbase = lf.AddIns.AddInBase()
        self._automation = lf.Automation.Automation(visible, List[String](""))
        self._application = self._automation.LightFieldApplication
        self._experiment = self._application.Experiment

    @property
    def automation(self):
        return self._automation

    @property
    def addinbase(self):
        return self._addinbase

    @property
    def application(self):
        return self._application

    @property
    def experiment(self):
        return self._experiment

    def __del__(self):
        """
        Uses Python garbage collection method used to terminate
        AddInProcess.exe, if this is not done, LightField will not reopen
        properly.
        """
        self.automation.Dispose()
        print('Closed AddInProcess.exe')

    def set(self, setting, value):
        """
        Helper function for setting experiment parameters with basic value
        checking.
        """

        if self.experiment.Exists(setting):
            if self.experiment.IsValid(setting, value):

                self.experiment.SetValue(setting, value)

            else:

                # TODO: add proper exception?
                print('Invalid value: {} for setting: {}.'.format(value, setting))

        else:

            # TODO: add proper exceptions?
            print('Invalid setting:{}'.format(setting))

        return

    def get(self, setting):
        """
        Helper function for getting experiment parameters with a basic check
        to see if a specific setting is valid.
        """

        if self.experiment.Exists(setting):

            value = self.experiment.GetValue(setting)

        else:

            value = []
            print('Invalid setting: {}'.format(setting))

        return value

    def load_experiment(self, value):
        self.experiment.Load(value)

    def acquire(self):
        """
        Helper function to acquire the data from the PIXIS-style cameras.
        Your mileage may vary for the Pylon - need to check array reorganization
        since there is only a single vertical pixel.

        For a single frame, the data is returned as a 2D array (just raw counts
        from each pixel). Taking a section in the vertical (400-pixel) direction
        corresponds to wavelength averaging.

        For multiple frames, each exposure is stacked in a third dimension, so
        averaging can be performed simply by summing over the different planes.
        """

        acquisition_time = self.get(lf.AddIns.CameraSettings.ShutterTimingExposureTime)
        num_frames =  self.get(lf.AddIns.ExperimentSettings.FrameSettingsFramesToStore)

        self.experiment.Acquire()

        sleep(0.001 * acquisition_time * num_frames) # waits exposure duration

        while self.experiment.IsRunning:
            sleep(0.1) # waits for experiment to finish

        last_file = self.application.FileManager.GetRecentlyAcquiredFileNames().GetItem(0)

        image_set = self.application.FileManager.OpenFile(last_file, FileAccess.Read)

        if image_set.Regions.Length == 1:

            if image_set.Frames == 1:

                frame = image_set.GetFrame(0, 0)
                data = np.reshape(np.fromiter(frame.GetData(),'uint16'), [frame.Width, frame.Height], order='F')

            else:
                data = np.array([])
                for i in range(0, image_set.Frames):

                    frame = image_set.GetFrame(0, i)
                    new_frame = np.fromiter(frame.GetData(), 'uint16')

                    new_frame = np.reshape(np.fromiter(frame.GetData(), 'uint16'), [frame.Width, frame.Height], order='F')
                    data = np.dstack((data, new_frame)) if data.size else new_frame


            return data


        else:
        # not sure when this situation actually arises, but I think multiple
        # regions of interest have to be set.
        #     data = cell(imageset.Regions.Length, 1)
            print('image_set.Regions not 1! this needs to be figured out!')
            print(image_set.Frames)
        #     for j in range(0, image_set.Regions.Length - 1):
        #
        #         if image_set.Frames == 1:
        #
        #             frame = image_set.GetFrame(j, 0)
        #             buf = np.reshape(np.fromiter(frame.GetData(), 'uint16'), frame.Width, frame.Height))
        #
        #         else:
        #
        #             buf = []
        #
        #             for i in range(0, image_set.Frames-1):
        #
        #                 frame = image_set.GetFrame(j, i)
        #                 buffer = np.dstack(buf, np.reshape(np.fromiter(frame.GetData(), 'uint16'), frame.Width, frame.Height, 1))
        #
        #         data[j+1] = buf

    def close(self):
        """
        """
        self.automation.Dispose()
        print('Closed AddInProcess.exe')


class Spectrometer(Driver):

    GRATINGS = ['[500nm,600][0][0]','[1.2um,300][1][0]','[500nm,150][2][0]']# TODO: make it pull gratings from SW directly

    def initialize(self):
        """
        Sets up LightField
        """
        self.lfm = LightFieldM(True)
        self.lfm.load_experiment('SpyreAutomation')

    def finalize(self):
        """
        Shuts the club downnnnn.
        """
        self.lfm.close()

    @Feat(units='nm')
    def center_wavelength(self):
        """
        Returns the center wavelength in nanometers.
        """
        return self.lfm.get(lf.AddIns.SpectrometerSettings.GratingCenterWavelength)

    def get_wavelengths(self):
        """
        Returns the wavelength calibration for a single frame.
        """
        size = self.lfm.experiment.SystemColumnCalibration.get_Length()

        result = np.empty(size)

        for _ in range(size):

            result[_] = self.lfm.experiment.SystemColumnCalibration[_]

        return result

    @center_wavelength.setter
    def center_wavelength(self, nanometers):
        """
        Sets the spectrometer center wavelength to nanometers.
        """
        # this avoids bug where if step and glue is selected, doesn't allow setting center wavelength
        self.lfm.set(lf.AddIns.ExperimentSettings.StepAndGlueEnabled, False)

        return self.lfm.set(lf.AddIns.SpectrometerSettings.GratingCenterWavelength, nanometers)

    @Feat()
    def grating(self):
        """
        Returns the current grating.
        """
        return self.lfm.get(lf.AddIns.SpectrometerSettings.GratingSelected)

    @grating.setter
    def grating(self, grating):
        """
        Sets the current grating to be the one specified by parameter grating.
        """
        # TODO: figure out the format for setting this

        print('figure out the format for this')

    @Feat()
    def gratings(self):
        """
        Returns a list of all installed gratings.
        """
        break_down = False

        if break_down:

            import re

            for g in GRATINGS:

                match = re.search(r"\[(\d+\.?\d+[nu]m),(\d+)\]\[(\d+)\]\[(\d+)\])",g)
                blaze, g_per_mm, slot, turret = match.groups()

        return GRATINGS

    @Feat()
    def num_frames(self):
        """
        Returns the number of frames taken during the acquisition.
        """
        return self.lfm.get(lf.AddIns.ExperimentSettings.FrameSettingsFramesToStore)

    @num_frames.setter
    def num_frames(self, num_frames):
        """
        Sets the number of frames to be taken during acquisition to number.
        """
        return self.lfm.set(lf.AddIns.ExperimentSettings.FrameSettingsFramesToStore, num_frames)

    @Feat()
    def exposure_time(self):
        """
        Returns the single frame exposure time (in ms).
        """
        return self.lfm.get(lf.AddIns.CameraSettings.ShutterTimingExposureTime)

    @exposure_time.setter
    def exposure_time(self, ms):
        """
        Sets the single frame exposure time to be ms (in milliseconds).
        """
        return self.lfm.set(lf.AddIns.CameraSettings.ShutterTimingExposureTime, ms)

    @Feat()
    def sensor_temperature(self):
        """
        Returns the current sensor temperature (in degrees Celsius).
        """
        return self.lfm.get(lf.AddIns.CameraSettings.SensorTemperatureReading)

    @Feat()
    def sensor_setpoint(self):
        """
        Returns the sensor setpoint temperature (in degrees Celsius).
        """
        return self.lfm.get(lf.AddIns.CameraSettings.SensorTemperatureSetPoint)

    @sensor_setpoint.setter
    def sensor_setpoint(self, deg_C):
        """
        Sets the sensor target temperature (in degrees Celsius) to deg_C.
        """
        return self.lfm.set(lf.AddIns.CameraSettings.SensorTemperatureSetPoint, deg_C)


    @Action()
    def acquire_frame(self):
        """
        Acquires a frame (or series of frames) from the spectrometer, and the
        corresponding wavelength data.
        """
        return self.lfm.acquire(), self.get_wavelengths()

    @Action()
    def acquire_step_and_glue(self, wavelength_range):
        """
        Acquires a step and glue (wavelength sweep) over the specified range.

        Wavelength range must have two elements (both in nm), corresponding
        to the starting and stopping wavelengths.

        Note that the wavelength must be calibrated for this to be meaningful!
        """

        lambda_min = wavelength_range[0]
        lambda_max = wavelength_range[1]

        try:

            self.lfm.set(lf.AddIns.ExperimentSettings.StepAndGlueEnabled, True)



        except:

            self.lfm.set(lf.AddIns.ExperimentSettings.StepAndGlueEnabled, False)
            print('Unable to perform step and glue, check settings.')

            return

        self.lfm.set(lf.AddIns.ExperimentSettings.StepAndGlueStartingWavelength, lambda_min)
        self.lfm.set(lf.AddIns.ExperimentSettings.StepAndGlueEndingWavelength, lambda_max)

        data = self.lfm.acquire()

        wavelength = np.linspace(lambda_min, lambda_max, data.shape[0])

        print('Wavelength data is not strictly correct, this just interpolates.')
        print('TODO: figure out how step and glue determines which wavelengths are used. This might be done in post processing.')
        print('If you want the true values, use the actual .spe file that is generated.')

        self.lfm.set(lf.AddIns.ExperimentSettings.StepAndGlueEnabled, False)

        return data, wavelength


    @Action()
    def acquire_background(self):
        """
        Acquires a background file and sets it to be used in the current scan.
        """
        # TODO: take background file, and save it in a location where it be accessed later

        print('not yet implemented')


def lantz_test():

    s = Spectrometer()
    s.initialize()

    print('Exposure time: {}ms'.format(s.exposure_time))
    s.exposure_time = 2.0
    print('Exposure time: {}ms'.format(s.exposure_time))
    s.exposure_time = 1.5
    print('Exposure time: {}ms'.format(s.exposure_time))

    print('Number of frames: {}'.format(s.num_frames))
    s.num_frames = 1
    print('Number of frames: {}'.format(s.num_frames))
    s.num_frames = 10
    print('Number of frames: {}'.format(s.num_frames))

    print('Sensor temperature setpoint: {} C'.format(s.sensor_setpoint))
    s.sensor_setpoint = -60.0
    print('Sensor temperature setpoint: {} C'.format(s.sensor_setpoint))
    s.sensor_setpoint = -70.0
    print('Sensor temperature setpoint: {} C'.format(s.sensor_setpoint))


    print('Sensor temperature: {} C'.format(s.sensor_temperature))

    print('Grating: {}'.format(s.grating))

    print('Center wavelength:{}'.format(s.center_wavelength))
    s.center_wavelength = 700.0
    print('Center wavelength:{}'.format(s.center_wavelength))
    s.center_wavelength = 730.0
    print('Center wavelength:{}'.format(s.center_wavelength))


    print('Acquring single frame')
    s.num_frames = 1
    data, wavelength = s.acquire_frame()
    print(data.shape)

    s.num_frames = 10
    print('Acquring 10 frames')
    data, wavelength = s.acquire_frame()
    print(data.shape)


    data, wavelength = s.acquire_step_and_glue([500.0, 950.0])

    print(data.shape, wavelength.shape)

    s.finalize()


def test():

    #import matplotlib.pyplot as plt

    lfm = LightFieldM(True)
    lfm.load_experiment('SpyreAutomation')

    lfm.set_frames(5)
    print(lfm.get(lf.AddIns.ExperimentSettings.FrameSettingsFramesToStore))
    lfm.set_frames(10)
    print(lfm.get(lf.AddIns.ExperimentSettings.FrameSettingsFramesToStore))
    lfm.set_frames(1)

    print(lfm.get(lf.AddIns.CameraSettings.ShutterTimingExposureTime))
    lfm.set_exposure(0.5)
    print(lfm.get(lf.AddIns.CameraSettings.ShutterTimingExposureTime))
    lfm.set_exposure(1.5)
    print(lfm.get(lf.AddIns.CameraSettings.ShutterTimingExposureTime))


    #data = lfm.acquire()
    #print(data)
    #plt.plot(data)
    #plt.show()

    lfm.set_frames(10)
    #data = lfm.acquire()
    #print(data)
    #plt.plot(data)
    #plt.show()

if __name__ == "__main__":
    #test()
    lantz_test()



from lantz.driver import Driver
from lantz import Feat, Action

from telnetlib import Telnet

class Bristol671(Driver):

    def __init__(self, ip, port=23, timeout=1):
        super().__init__()
        self.con_args = [ip,port,timeout]

    def initialize(self):
        self.resource = Telnet(*self.con_args)
        self.clear_read_buffer()
    
    def clear_read_buffer(self, timeout=0.1):
        while self.read(timeout=timeout) != '':
            pass

    def finalize(self):
        self.resource.close()

    def write(self, cmd, write_termination='\r\n'):
        return self.resource.write(bytes(cmd+write_termination, 'ascii'))

    def read(self, read_termination='\r\n', timeout=1):
        byte_ans = self.resource.read_until(bytes(read_termination, 'ascii'), timeout)
        ans = byte_ans.decode('ascii')
        return ans.strip(read_termination)

    def query(self, cmd, write_termination='\r\n', read_termination='\r\n'):
        self.write(cmd, write_termination=write_termination)
        return self.read(read_termination=read_termination)

    @Feat()
    def idn(self):
        return self.query('*IDN?')

    @Feat(units='nm')
    def wavelength(self):
        return float(self.query(':MEAS:WAV?'))

    @Feat(units='THz')
    def frequency(self):
        return float(self.query(':MEAS:FREQ?'))

    @Feat(units='mW')
    def power(self):
        self.write(':UNIT:POW MW')
        return float(self.query(':MEAS:POW?'))



# some stuff
"""

    lantz.drivers.agilent.AG33522a
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Implementation of 33522a AWG with 2 channels

    The ascii commands: https://ecee.colorado.edu/~mathys/ecen1400/pdf/references/Agilent33500_FunctionGenerator.pdf

    Author: Berk Diler
    
    Date: 8/16/2017
    
"""


import numpy as np
import lantz
from lantz import Action, Feat, DictFeat, ureg
from collections import OrderedDict
from lantz.messagebased import MessageBasedDriver

class AG33522A(MessageBasedDriver):
    
    DEFAULTS = {
        'COMMON': {
            'write_termination': '\n',
            'read_termination': '\n',
        }
    }

    CHANNELS = OrderedDict([(1, 1),
                           (2,2)])

    TOGGLE = OrderedDict([("ON", 1),
                          ("OFF", 0)])

    WAVEFORMS = OrderedDict([('sine', "SIN"),
                             ('square', "SQU"),
                             ('triangle', "TRI"),
                             ('ramp', "RAMP"),
                             ('pulse', 'PULS'),
                             ('ps random noise', "PRBS"),
                             ('noise', 'NOIS'),
                             ('arbitrary', 'ARB'),
                             ('dc', "DC")])
    
    @Feat(read_once=True)
    def idn(self):
        return self.query('*IDN?')

    @Action()
    def reset(self):
        self.write("*RST")

    @Feat()
    def error(self):
        answ = self.query("SYST:ERR?")
        if answ != '+0,"No error"':
            return answ
        else:
            return

    @DictFeat(keys = CHANNELS, units="Hz", limits=(1e-6, 20e6))
    def frequency(self,channel):
        """
        Frequency of the channel
        """
        return float(self.query('SOUR{}:FREQ?'.format(channel)))

    @frequency.setter
    def frequency(self,channel,value):
        self.write('SOUR{}:FREQ {:1.6f}'.format(channel,value))

    @DictFeat(keys = CHANNELS, values = TOGGLE)
    def output(self,channel):
        """
        Output of the channel
        """
        return float(self.query('OUTP{}?'.format(channel)))

    @output.setter
    def output(self,channel,val):
        self.write('OUTP{} {}'.format(channel,val))

    @DictFeat(keys = CHANNELS, units = "V", limits=(-10., 10.))
    def voltage(self,channel):
        """
        Output voltage of the channel
        """
        return float(self.query("SOUR{}:VOLT?".format(channel)))

    @voltage.setter
    def voltage(self,channel,value):
        self.write("SOUR{}:VOLT {:1.6f}".format(channel,value))

    @DictFeat(keys=CHANNELS, units="V", limits=(-10., 10.))
    def offset(self, channel):
        """
        Output voltage of the channel
        """
        return float(self.query("SOUR{}:VOLT:OFFS?".format(channel)))

    @offset.setter
    def offset(self, channel, value):
        self.write("SOUR{}:VOLT:OFFS {:1.6f}".format(channel, value))

    @DictFeat(keys=CHANNELS, units="V", limits=(-10.,10.))
    def lower_limit(self,channel):
        """
        Lower voltage limit of the channel
        """
        return float(self.query("SOUR{}:VOLT:LIM:LOW?".format(channel)))

    @lower_limit.setter
    def lower_limit(self,channel,value):
        self.write("SOUR{}:VOLT:LIM:LOW {:1.6f}".format(channel, value))

    @DictFeat(keys=CHANNELS, units="V", limits=(-10.,10.))
    def high_limit(self,channel):
        """
        High voltage limit of the channel
        """
        return float(self.query("SOUR{}:VOLT:LIM:HIGH?".format(channel)))

    @high_limit.setter
    def high_limit(self,channel,value):
        self.write("SOUR{}:VOLT:LIM:HIGH {:1.6f}".format(channel, value))

    @DictFeat(keys = CHANNELS, values=TOGGLE)
    def limit(self,channel):
        """
        limit of the channel
        """
        return float(self.query('SOUR{}:VOLT:LIM:STAT?'.format(channel)))

    @limit.setter
    def limit(self,channel, value):
        self.write('SOUR{}:VOLT:LIM:STAT {}'.format(channel,value))

    @DictFeat(keys=CHANNELS, values=WAVEFORMS)
    def function(self,channel):
        """
        Function of the channel
        """
        return self.query('SOUR{}:FUNC?'.format(channel))

    @function.setter
    def function(self,channel,value):
        self.write('SOUR{}:FUNC {}'.format(channel, value))


# -*- coding: utf-8 -*-
"""
    lantz.drivers.princetoninstruments.winspec
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Implementation of Winspec PI Camera over a Socket
    Author: Berk Diler
    Date: 29/08/2017
"""

from collections import OrderedDict
from lantz import Action, Feat, DictFeat, Q_
from lantz.messagebased import MessageBasedDriver
import json
import numpy as np



class Winspec(MessageBasedDriver):

    DEFAULTS = {
        'COMMON': {
            'write_termination': '',
            'read_termination': '\r\n',
        }
    }

    GRATINGS = OrderedDict([
        (1,1),
        (2,2)
    ])

    CALIBRATION_PARAMS = OrderedDict([
        (1, {
            "d": 1./600.,
            "gamma": 0.479941,
            "fl": 304.223,
            "delta": -0.00882795,
            "dpixel": 2.33503,
            "dl1": -0.0446331,
            "dl2": -0.0000352349
        }),
        (2, {
            "d": 1. / 150.,
            "gamma": -0.0101195,
            "fl": 320.391,
            "delta": 0.00946824,
            "dpixel": 2.6969,
            "dl1": -0.0308197,
            "dl2": 0.0000550113
        }),
        ("constants", {
            "pwidth" : 25.e-3,
            "center_pixel": 512.5,
            "m": 1
        })
    ])

    @Feat(units="s")
    def exposure_time(self):
        return float(self.query("get exposure time"))

    @exposure_time.setter
    def exposure_time(self, exp_time):
        if self.query("set exposure time {:1.3e}".format(exp_time)) != "OK":
            raise Exception

    @Feat(units="nm")
    def wavelength(self):
        return float(self.query("get wavelength"))

    @wavelength.setter
    def wavelength(self, wl):
        if self.query("set wavelength {:1.3e}".format(wl)) != "OK":
            raise Exception
        
    @Feat(values=GRATINGS)
    def grating(self):
        return int(self.query("get grating"))

    @grating.setter
    def grating(self, wl):
        if self.query("set grating {:1.3e}".format(wl)) != "OK":
            raise Exception
    
    @Feat(units="kelvin", limits=(174,248))
    def target_temperature(self):
        temp_c = float(self.query("get target temperature"))
        result = Q_(temp_c, "degC").to("kelvin").magnitude
        return result

    @target_temperature.setter
    def target_temperature(self, t):
        t = Q_(t, "kelvin").to("degC").magnitude
        if self.query("set target temperature {:1.3e}".format(t)) != "OK":
            raise Exception

    @Feat(units="kelvin")
    def temperature(self):
        temp_c = float(self.query("get temperature"))
        result = Q_(temp_c, "degC").to("kelvin").magnitude
        return result

    @DictFeat(read_once=True, keys=GRATINGS)
    def grating_grooves(self, grating):
        return self.query("get grating grooves {}".format(grating))

    @Action()
    def get_exposure(self):
        return np.array(json.loads(self.query("get spectrum")))

    @Action()
    def get_int_wavelength(self):
        return np.array(json.loads(self.query("get wavelengths")))

    @Action()
    def get_cal_wavelength(self):
        grating = self.grating
        center_wl = self.wavelength.magnitude
        d = self.CALIBRATION_PARAMS[grating]["d"]
        gamma = self.CALIBRATION_PARAMS[grating]["gamma"]
        fl = self.CALIBRATION_PARAMS[grating]["fl"]
        delta = self.CALIBRATION_PARAMS[grating]["delta"]
        dpixel = self.CALIBRATION_PARAMS[grating]["dpixel"]
        dl1 = self.CALIBRATION_PARAMS[grating]["dl1"]
        dl2 = self.CALIBRATION_PARAMS[grating]["dl2"]
        pwidth = self.CALIBRATION_PARAMS["constants"]["pwidth"]
        center_pixel = self.CALIBRATION_PARAMS["constants"]["center_pixel"]
        m = self.CALIBRATION_PARAMS["constants"]["m"]

        x_array = np.array(range(1024)) + 1
        x_array = x_array - center_pixel + dpixel + dl1 * (center_wl - 1000.0) + dl2 * (center_wl - 1000.0) * (
        center_wl - 1000.0)
        xi = np.arctan(x_array * pwidth * np.cos(delta) / (fl + x_array * pwidth * np.sin(delta)))
        psi = np.arcsin(m * center_wl * 1.0e-6 / (2.0 * d * np.cos(gamma / 2.0)))
        wl = (d / m) * (np.sin(psi - gamma / 2.0) + np.sin(psi + gamma / 2.0 + xi)) * 1.0e6
        return wl



# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.usbtmc
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements a USBDriver based class to control USBTMC instruments

    Loosely based on PyUSBTMC:python module to handle USB-TMC(Test and Measurement class)　devices.
    by Noboru Yamamot, Accl. Lab, KEK, JAPAN

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import enum
import time
import struct
from collections import namedtuple

import usb
from lantz.drivers.legacy.textual import TextualMixin
from lantz import errors
from lantz.drivers.legacy.usb import find_devices, find_interfaces, find_endpoint, USBDriver


class MSGID(enum.IntEnum):
    """From USB-TMC table2
    """
    DEV_DEP_MSG_OUT = 1
    REQUEST_DEV_DEP_MSG_IN = 2
    DEV_DEP_MSG_IN = 2
    VENDOR_SPECIFIC_OUT = 126
    REQUEST_VENDOR_SPECIFIC_IN = 127
    VENDOR_SPECIFIC_IN = 127
    TRIGGER = 128 # for USB488


class REQUEST(enum.IntEnum):
    INITIATE_ABORT_BULK_OUT = 1
    CHECK_ABORT_BULK_OUT_STATUS = 2
    INITIATE_ABORT_BULK_IN = 3
    CHECK_ABORT_BULK_IN_STATUS = 4
    INITIATE_CLEAR = 5
    CHECK_CLEAR_STATUS = 6
    GET_CAPABILITIES = 7
    INDICATOR_PULSE = 64


def find_tmc_devices(vendor=None, product=None, serial_number=None, custom_match=None, **kwargs):
    """Find connected USBTMC devices. See lantz.usb.find_devices for more info.
    """
    def is_usbtmc(dev):
        if custom_match and not custom_match(dev):
            return False
        return bool(find_interfaces(dev, bInterfaceClass=0xfe, bInterfaceSubClass=3))

    return find_devices(vendor, product, serial_number, is_usbtmc, **kwargs)


class BulkOutMessage(object):
    """The Host uses the Bulk-OUT endpoint to send USBTMC command messages to the device.
    """

    @staticmethod
    def build_array(btag, eom, chunk):
        size = len(chunk)
        return struct.pack('BBBx', MSGID.DEV_DEP_MSG_OUT, btag, ~btag & 0xFF) + \
               struct.pack("<LBxxx", size, eom) + \
               chunk + \
               b'\0' * ((4 - size) % 4)


class BulkInMessage(namedtuple('BulkInMessage', 'msgid btag btaginverse '
                                                'transfer_size transfer_attributes data')):
    """The Host uses the Bulk-IN endpoint to read USBTMC response messages from the device.
    The Host must first send a USBTMC command message that expects a response before
    attempting to read a USBTMC response message.
    """

    @classmethod
    def from_bytes(cls, data):
        msgid, btag, btaginverse = struct.unpack_from('BBBx', data)
        assert msgid == MSGID.DEV_DEP_MSG_IN

        transfer_size, transfer_attributes = struct.unpack_from('<LBxxx', data, 4)

        data = data[12:]
        return cls(msgid, btag, btaginverse, transfer_size, transfer_attributes, data)


    @staticmethod
    def build_array(btag, transfer_size, term_char=None):
        """

        :param transfer_size:
        :param btag:
        :param term_char:
        :return:
        """

        if term_char is None:
            transfer_attributes = 0
            term_char = 0
        else:
            transfer_attributes = 2

        return struct.pack('BBBx', MSGID.REQUEST_DEV_DEP_MSG_IN, btag, ~btag & 0xFF) + \
               struct.pack("<LBBxx", transfer_size, transfer_attributes, term_char)


class USBTMCDriver(USBDriver, TextualMixin):

    SEND_TERMINATION = ''
    RECV_TERMINATION = '\n'

    RECV_CHUNK = 1024 ** 2

    find_devices = staticmethod(find_tmc_devices)

    def __init__(self, vendor=None, product=None, serial_number=None, **kwargs):
        super().__init__(vendor, product, serial_number, **kwargs)
        self.usb_intr_in = find_endpoint(self.usb_intf, usb.ENDPOINT_IN, usb.ENDPOINT_TYPE_INTERRUPT)
        self.log_debug('EP Address: intr={}'.format(self.usb_intr_in.bEndpointAddress))

        self.usb_dev.reset()

        time.sleep(0.01)

        self._get_capabilities()
    
        self._btag = 0

        if not (self.usb_recv_ep and self.usb_send_ep):
            raise errors.InstrumentError("TMC device must have both Bulk-In and Bulk-out endpoints.")

    def _get_capabilities(self):
        cap = self.usb_dev.ctrl_transfer(
                   usb.util.build_request_type(usb.util.CTRL_IN,
                                               usb.util.CTRL_TYPE_CLASS,
                                               usb.util.CTRL_RECIPIENT_INTERFACE),
                   REQUEST.GET_CAPABILITIES,
                   0x0000,
                   self.usb_intf.index,
                   0x0018,
                   timeout=self.TIMEOUT)

    def _find_interface(self, dev, setting):
        interfaces = find_interfaces(dev, bInterfaceClass=0xFE, bInterfaceSubClass=3)
        if not interfaces:
            raise errors.InstrumentError('USB TMC interface not found.')
        elif len(interfaces) > 1:
            self.log_warning('More than one interface found, selecting first.')

        return interfaces[0]

    def send(self, command, termination=None, encoding=None):
        if termination is None:
            termination = self.SEND_TERMINATION
        if encoding is None:
            encoding = self.ENCODING

        message = bytes(command + termination, encoding)
        self.log_debug('Sending {}', message)

        begin, end, size = 0, 0, len(message)
        bytes_sent = 0
        while not end > size:
            begin, end = end, begin + self.RECV_CHUNK

            self._btag = (self._btag % 255) + 1

            data = BulkOutMessage.build_array(self._btag, end > size, message[begin:end])

            bytes_sent += self.raw_send(data)

        return bytes_sent

    send.__doc__ = TextualMixin.send.__doc__

    def recv(self, termination=None, encoding=None, recv_chunk=None):

        termination = termination or self.RECV_TERMINATION
        encoding = encoding or self.ENCODING
        recv_chunk = recv_chunk or self.RECV_CHUNK

        eom = False

        received = self._received

        while not (termination in received or eom):
            self._btag = (self._btag % 255) + 1

            req = BulkInMessage.build_array(self._btag, recv_chunk, None)

            self.raw_send(req)

            resp = self.raw_recv(recv_chunk)

            response = BulkInMessage.from_bytes(resp)

            received += str(response.data, encoding)
            eom = response.transfer_attributes & 1

        self.log_debug('Received {!r} (len={})', received, len(received))

        received, self._received = received.split(termination, 1)

        return received

    recv.__doc__ = TextualMixin.recv.__doc__




# -*- coding: utf-8 -*-
"""
    lantz.utils.rpc
    ~~~~~~~~~~~~~~~

    Sun RPC version 2 -- RFC1057

    This file is drawn from Python's RPC demo, updated for python 3.

    XXX There should be separate exceptions for the various reasons why
    XXX an RPC can fail, rather than using RuntimeError for everything

    XXX The UDP version of the protocol resends requests when it does
    XXX not receive a timely reply -- use only for idempotent calls!

    XXX There is no provision for call timeout on TCP connections

    Original source: http://svn.python.org/projects/python/trunk/Demo/rpc/rpc.py

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


import sys
import enum
import xdrlib
import socket
import struct

#: Version of the protocol
RPCVERSION = 2


class msg_type(enum.IntEnum):
    CALL = 0
    REPLY = 1


class auth_flavor(enum.IntEnum):
    AUTH_NULL = 0
    AUTH_UNIX = 1
    AUTH_SHORT = 2
    AUTH_DES = 3


class reply_stat(enum.IntEnum):
    MSG_ACCEPTED = 0
    MSG_DENIED = 1


class accept_stat(enum.IntEnum):
    SUCCESS = 0            # RPC executed successfully
    PROG_UNAVAIL = 1      # remote hasn't exported program
    PROG_MISMATCH = 2      # remote can't support version #
    PROC_UNAVAIL = 3      # program can't support procedure
    GARBAGE_ARGS = 4      # procedure can't decode params


class reject_stat(enum.IntEnum):
    RPC_MISMATCH = 0       # RPC version number != 2
    AUTH_ERROR = 1         # remote can't authenticate caller


class auth_stat(enum.IntEnum):
    AUTH_OK = 0
    AUTH_BADCRED = 1  # bad credentials (seal broken)
    AUTH_REJECTEDCRED = 2  # client must begin new session
    AUTH_BADVERF = 3  # bad verifier (seal broken)
    AUTH_REJECTEDVERF = 4  # verifier expired or replayed
    AUTH_TOOWEAK = 5  # rejected for security reasons


# Exceptions
class RPCError(Exception):
    pass

class RPCBadFormat(RPCError):
    pass

class RPCBadVersion(RPCError):
    pass

class RPCGarbageArgs(RPCError):
    pass

class RPCUnpackError(RPCError):
    pass


def make_auth_null():
    return b''


class Packer(xdrlib.Packer):

    def pack_auth(self, auth):
        flavor, stuff = auth
        self.pack_enum(flavor)
        self.pack_opaque(stuff)

    def pack_auth_unix(self, stamp, machinename, uid, gid, gids):
        self.pack_uint(stamp)
        self.pack_string(machinename)
        self.pack_uint(uid)
        self.pack_uint(gid)
        self.pack_uint(len(gids))
        for i in gids:
            self.pack_uint(i)

    def pack_callheader(self, xid, prog, vers, proc, cred, verf):
        self.pack_uint(xid)
        self.pack_enum(msg_type.CALL)
        self.pack_uint(RPCVERSION)
        self.pack_uint(prog)
        self.pack_uint(vers)
        self.pack_uint(proc)
        self.pack_auth(cred)
        self.pack_auth(verf)
        # Caller must add procedure-specific part of call

    def pack_replyheader(self, xid, verf):
        self.pack_uint(xid)
        self.pack_enum(msg_type.REPLY)
        self.pack_uint(reply_stat.MSG_ACCEPTED)
        self.pack_auth(verf)
        self.pack_enum(accept_stat.SUCCESS)
        # Caller must add procedure-specific part of reply


class Unpacker(xdrlib.Unpacker):

    def unpack_auth(self):
        flavor = self.unpack_enum()
        stuff = self.unpack_opaque()
        return (flavor, stuff)

    def unpack_callheader(self):
        xid = self.unpack_uint()
        temp = self.unpack_enum()
        if temp != msg_type.CALL:
            raise RPCBadFormat('no CALL but %r' % (temp,))
        temp = self.unpack_uint()
        if temp != RPCVERSION:
            raise RPCBadVersion('bad RPC version %r' % (temp,))
        prog = self.unpack_uint()
        vers = self.unpack_uint()
        proc = self.unpack_uint()
        cred = self.unpack_auth()
        verf = self.unpack_auth()
        return xid, prog, vers, proc, cred, verf
        # Caller must add procedure-specific part of call

    def unpack_replyheader(self):
        xid = self.unpack_uint()
        mtype = self.unpack_enum()
        if mtype != msg_type.REPLY:
            raise RPCUnpackError('no REPLY but %r' % (mtype,))
        stat = self.unpack_enum()
        if stat == reply_stat.MSG_DENIED:
            stat = self.unpack_enum()
            if stat == reject_stat.RPC_MISMATCH:
                low = self.unpack_uint()
                high = self.unpack_uint()
                raise RPCUnpackError('MSG_DENIED: RPC_MISMATCH: %r' % ((low, high),))
            if stat == reject_stat.AUTH_ERROR:
                stat = self.unpack_uint()
                raise RPCUnpackError('MSG_DENIED: AUTH_ERROR: %r' % (stat,))
            raise RPCUnpackError('MSG_DENIED: %r' % (stat,))
        if stat != reply_stat.MSG_ACCEPTED:
            raise RPCUnpackError('Neither MSG_DENIED nor MSG_ACCEPTED: %r' % (stat,))
        verf = self.unpack_auth()
        stat = self.unpack_enum()
        if stat == accept_stat.PROG_UNAVAIL:
            raise RPCUnpackError('call failed: PROG_UNAVAIL')
        if stat == accept_stat.PROG_MISMATCH:
            low = self.unpack_uint()
            high = self.unpack_uint()
            raise RPCUnpackError('call failed: PROG_MISMATCH: %r' % ((low, high),))
        if stat == accept_stat.PROC_UNAVAIL:
            raise RPCUnpackError('call failed: PROC_UNAVAIL')
        if stat == accept_stat.GARBAGE_ARGS:
            raise RPCGarbageArgs
        if stat != accept_stat.SUCCESS:
            raise RPCUnpackError('call failed: %r' % (stat,))
        return xid, verf
        # Caller must get procedure-specific part of reply


class Client:
    """Common base class for clients.
    """

    def __init__(self, host, prog, vers, port):
        self.host = host
        self.prog = prog
        self.vers = vers
        self.port = port
        self.lastxid = 0 # XXX should be more random?
        self.cred = None
        self.verf = None

    def make_call(self, proc, args, pack_func, unpack_func):
        # Don't normally override this (but see Broadcast)
        if pack_func is None and args is not None:
            raise TypeError('non-null args with null pack_func')
        self.start_call(proc)
        if pack_func:
            pack_func(args)
        self.do_call()
        if unpack_func:
            result = unpack_func()
        else:
            result = None
        self.unpacker.done()
        return result

    def start_call(self, proc):
        # Don't override this
        self.lastxid = xid = self.lastxid + 1
        cred = self.mkcred()
        verf = self.mkverf()
        p = self.packer
        p.reset()
        p.pack_callheader(xid, self.prog, self.vers, proc, cred, verf)

    def do_call(self):
        # This MUST be overridden
        raise RPCError('do_call not defined')

    def mkcred(self):
        # Override this to use more powerful credentials
        if self.cred is None:
            self.cred = (auth_flavor.AUTH_NULL, make_auth_null())
        return self.cred

    def mkverf(self):
        # Override this to use a more powerful verifier
        if self.verf is None:
            self.verf = (auth_flavor.AUTH_NULL, make_auth_null())
        return self.verf

    def call_0(self):
        # Procedure 0 is always like this
        return self.make_call(0, None, None, None)


# Record-Marking standard support

def sendfrag(sock, last, frag):
    x = len(frag)
    if last:
        x = x | 0x80000000
    header = struct.pack(">I", x)
    sock.send(header + frag)


def sendrecord(sock, record):
    sendfrag(sock, 1, record)


def recvfrag(sock):
    header = sock.recv(4)
    if len(header) < 4:
        raise EOFError
    x = struct.unpack(">I", header[0:4])[0]
    last = ((x & 0x80000000) != 0)
    n = int(x & 0x7fffffff)
    frag = b''
    while n > 0:
        buf = sock.recv(n)
        if not buf:
            raise EOFError
        n = n - len(buf)
        frag = frag + buf
    return last, frag


def recvrecord(sock):
    record = b''
    last = 0
    while not last:
        last, frag = recvfrag(sock)
        record = record + frag
    return record


class RawTCPClient(Client):
    """Client using TCP to a specific port.
    """
    def __init__(self, host, prog, vers, port):
        Client.__init__(self, host, prog, vers, port)
        self.connect()
    
    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.host, self.port))
        
    def close(self):
        self.sock.close()
    
    def do_call(self):
        call = self.packer.get_buf()
        sendrecord(self.sock, call)
        reply = recvrecord(self.sock)
        u = self.unpacker
        u.reset(reply)
        xid, verf = u.unpack_replyheader()
        if xid != self.lastxid:
            # Can't really happen since this is TCP...
            raise RPCError('wrong xid in reply %r instead of %r' % (xid, self.lastxid))


# Client using UDP to a specific port

class RawUDPClient(Client):
    def __init__(self, host, prog, vers, port):
        Client.__init__(self, host, prog, vers, port)
        self.connect()
    
    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.connect((self.host, self.port))
        
    def close(self):
        self.sock.close()

    def do_call(self):
        call = self.packer.get_buf()
        self.sock.send(call)
        try:
            from select import select
        except ImportError:
            print('WARNING: select not found, RPC may hang')
            select = None
        BUFSIZE = 8192 # Max UDP buffer size
        timeout = 1
        count = 5
        while 1:
            r, w, x = [self.sock], [], []
            if select:
                r, w, x = select(r, w, x, timeout)
            if self.sock not in r:
                count = count - 1
                if count < 0: raise RPCError('timeout')
                if timeout < 25:
                    timeout = timeout *2
                self.sock.send(call)
                continue
            reply = self.sock.recv(BUFSIZE)
            u = self.unpacker
            u.reset(reply)
            xid, verf = u.unpack_replyheader()
            if xid != self.lastxid:
                continue
            break


class RawBroadcastUDPClient(RawUDPClient):
    """Client using UDP broadcast to a specific port.
    """

    def __init__(self, bcastaddr, prog, vers, port):
        RawUDPClient.__init__(self, bcastaddr, prog, vers, port)
        self.reply_handler = None
        self.timeout = 30
    
    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    def set_reply_handler(self, reply_handler):
        self.reply_handler = reply_handler

    def set_timeout(self, timeout):
        self.timeout = timeout # Use None for infinite timeout

    def make_call(self, proc, args, pack_func, unpack_func):
        if pack_func is None and args is not None:
            raise TypeError('non-null args with null pack_func')
        self.start_call(proc)
        if pack_func:
            pack_func(args)
        call = self.packer.get_buf()
        self.sock.sendto(call, (self.host, self.port))
        try:
            from select import select
        except ImportError:
            print('WARNING: select not found, broadcast will hang')
            select = None
        BUFSIZE = 8192 # Max UDP buffer size (for reply)
        replies = []
        if unpack_func is None:
            def dummy(): pass
            unpack_func = dummy
        while 1:
            r, w, x = [self.sock], [], []
            if select:
                if self.timeout is None:
                    r, w, x = select(r, w, x)
                else:
                    r, w, x = select(r, w, x, self.timeout)
            if self.sock not in r:
                break
            reply, fromaddr = self.sock.recvfrom(BUFSIZE)
            u = self.unpacker
            u.reset(reply)
            xid, verf = u.unpack_replyheader()
            if xid != self.lastxid:
                continue
            reply = unpack_func()
            self.unpacker.done()
            replies.append((reply, fromaddr))
            if self.reply_handler:
                self.reply_handler(reply, fromaddr)
        return replies


# Port mapper interface

# Program number, version and port number
PMAP_PROG = 100000
PMAP_VERS = 2
PMAP_PORT = 111

class PMAP_VERS(enum.IntEnum):
    PMAPPROC_NULL = 0                       # (void) -> void
    PMAPPROC_SET = 1                        # (mapping) -> bool
    PMAPPROC_UNSET = 2                      # (mapping) -> bool
    PMAPPROC_GETPORT = 3                    # (mapping) -> unsigned int
    PMAPPROC_DUMP = 4                       # (void) -> pmaplist
    PMAPPROC_CALLIT = 5                     # (call_args) -> call_result

# A mapping is (prog, vers, prot, port) and prot is one of:

IPPROTO_TCP = 6
IPPROTO_UDP = 17

# A pmaplist is a variable-length list of mappings, as follows:
# either (1, mapping, pmaplist) or (0).

# A call_args is (prog, vers, proc, args) where args is opaque;
# a call_result is (port, res) where res is opaque.


class PortMapperPacker(Packer):

    def pack_mapping(self, mapping):
        prog, vers, prot, port = mapping
        self.pack_uint(prog)
        self.pack_uint(vers)
        self.pack_uint(prot)
        self.pack_uint(port)

    def pack_pmaplist(self, list):
        self.pack_list(list, self.pack_mapping)

    def pack_call_args(self, ca):
        prog, vers, proc, args = ca
        self.pack_uint(prog)
        self.pack_uint(vers)
        self.pack_uint(proc)
        self.pack_opaque(args)


class PortMapperUnpacker(Unpacker):

    def unpack_mapping(self):
        prog = self.unpack_uint()
        vers = self.unpack_uint()
        prot = self.unpack_uint()
        port = self.unpack_uint()
        return prog, vers, prot, port

    def unpack_pmaplist(self):
        return self.unpack_list(self.unpack_mapping)

    def unpack_call_result(self):
        port = self.unpack_uint()
        res = self.unpack_opaque()
        return port, res


class PartialPortMapperClient:

    def __init__(self):
        self.packer = PortMapperPacker()
        self.unpacker = PortMapperUnpacker('')

    def set(self, mapping):
        return self.make_call(PMAP_VERS.PMAPPROC_SET, mapping,
                              self.packer.pack_mapping,
                              self.unpacker.unpack_uint)

    def unset(self, mapping):
        return self.make_call(PMAP_VERS.PMAPPROC_UNSET, mapping,
                              self.packer.pack_mapping,
                              self.unpacker.unpack_uint)

    def get_port(self, mapping):
        return self.make_call(PMAP_VERS.PMAPPROC_GETPORT, mapping,
                              self.packer.pack_mapping,
                              self.unpacker.unpack_uint)

    def dump(self):
        return self.make_call(PMAP_VERS.PMAPPROC_DUMP, None,
                              None,
                              self.unpacker.unpack_pmaplist)

    def callit(self, ca):
        return self.make_call(PMAP_VERS.PMAPPROC_CALLIT, ca,
                              self.packer.pack_call_args,
                              self.unpacker.unpack_call_result)


class TCPPortMapperClient(PartialPortMapperClient, RawTCPClient):

    def __init__(self, host):
        RawTCPClient.__init__(self, host, PMAP_PROG, PMAP_VERS, PMAP_PORT)
        PartialPortMapperClient.__init__(self)


class UDPPortMapperClient(PartialPortMapperClient, RawUDPClient):

    def __init__(self, host):
        RawUDPClient.__init__(self, host, PMAP_PROG, PMAP_VERS, PMAP_PORT)
        PartialPortMapperClient.__init__(self)


class BroadcastUDPPortMapperClient(PartialPortMapperClient, RawBroadcastUDPClient):

    def __init__(self, bcastaddr):
        RawBroadcastUDPClient.__init__(self, bcastaddr, PMAP_PROG, PMAP_VERS, PMAP_PORT)
        PartialPortMapperClient.__init__(self)


class TCPClient(RawTCPClient):
    """A TCP Client that find their server through the Port mapper
    """
    def __init__(self, host, prog, vers):
        pmap = TCPPortMapperClient(host)
        port = pmap.get_port((prog, vers, IPPROTO_TCP, 0))
        pmap.close()
        if port == 0:
            raise RPCError('program not registered')
        RawTCPClient.__init__(self, host, prog, vers, port)


class UDPClient(RawUDPClient):
    """A UDP Client that find their server through the Port mapper
    """
    def __init__(self, host, prog, vers):
        pmap = UDPPortMapperClient(host)
        port = pmap.get_port((prog, vers, IPPROTO_UDP, 0))
        pmap.close()
        if port == 0:
            raise RPCError('program not registered')
        RawUDPClient.__init__(self, host, prog, vers, port)


class BroadcastUDPClient(Client):
    """A Broadcast UDP Client that find their server through the Port mapper
    """

    def __init__(self, bcastaddr, prog, vers):
        self.pmap = BroadcastUDPPortMapperClient(bcastaddr)
        self.pmap.set_reply_handler(self.my_reply_handler)
        self.prog = prog
        self.vers = vers
        self.user_reply_handler = None
        self.addpackers()

    def close(self):
        self.pmap.close()

    def set_reply_handler(self, reply_handler):
        self.user_reply_handler = reply_handler

    def set_timeout(self, timeout):
        self.pmap.set_timeout(timeout)

    def my_reply_handler(self, reply, fromaddr):
        port, res = reply
        self.unpacker.reset(res)
        result = self.unpack_func()
        self.unpacker.done()
        self.replies.append((result, fromaddr))
        if self.user_reply_handler is not None:
            self.user_reply_handler(result, fromaddr)

    def make_call(self, proc, args, pack_func, unpack_func):
        self.packer.reset()
        if pack_func:
            pack_func(args)
        if unpack_func is None:
            def dummy(): pass
            self.unpack_func = dummy
        else:
            self.unpack_func = unpack_func
        self.replies = []
        packed_args = self.packer.get_buf()
        dummy_replies = self.pmap.Callit((self.prog, self.vers, proc, packed_args))
        return self.replies


# Server classes

# These are not symmetric to the Client classes
# XXX No attempt is made to provide authorization hooks yet

class Server:

    def __init__(self, host, prog, vers, port):
        self.host = host # Should normally be '' for default interface
        self.prog = prog
        self.vers = vers
        self.port = port # Should normally be 0 for random port
        self.port = port
        self.addpackers()

    def register(self):
        mapping = self.prog, self.vers, self.prot, self.port
        p = TCPPortMapperClient(self.host)
        if not p.set(mapping):
            raise RPCError('register failed')

    def unregister(self):
        mapping = self.prog, self.vers, self.prot, self.port
        p = TCPPortMapperClient(self.host)
        if not p.unset(mapping):
            raise RPCError('unregister failed')

    def handle(self, call):
        # Don't use unpack_header but parse the header piecewise
        # XXX I have no idea if I am using the right error responses!
        self.unpacker.reset(call)
        self.packer.reset()
        xid = self.unpacker.unpack_uint()
        self.packer.pack_uint(xid)
        temp = self.unpacker.unpack_enum()
        if temp != msg_type.CALL:
            return None # Not worthy of a reply
        self.packer.pack_uint(msg_type.REPLY)
        temp = self.unpacker.unpack_uint()
        if temp != RPCVERSION:
            self.packer.pack_uint(reply_stat.MSG_DENIED)
            self.packer.pack_uint(reject_stat.RPC_MISMATCH)
            self.packer.pack_uint(RPCVERSION)
            self.packer.pack_uint(RPCVERSION)
            return self.packer.get_buf()
        self.packer.pack_uint(reply_stat.MSG_ACCEPTED)
        self.packer.pack_auth((auth_flavor.AUTH_NULL, make_auth_null()))
        prog = self.unpacker.unpack_uint()
        if prog != self.prog:
            self.packer.pack_uint(accept_stat.PROG_UNAVAIL)
            return self.packer.get_buf()
        vers = self.unpacker.unpack_uint()
        if vers != self.vers:
            self.packer.pack_uint(accept_stat.PROG_MISMATCH)
            self.packer.pack_uint(self.vers)
            self.packer.pack_uint(self.vers)
            return self.packer.get_buf()
        proc = self.unpacker.unpack_uint()
        methname = 'handle_' + repr(proc)
        try:
            meth = getattr(self, methname)
        except AttributeError:
            self.packer.pack_uint(accept_stat.PROC_UNAVAIL)
            return self.packer.get_buf()
        cred = self.unpacker.unpack_auth()
        verf = self.unpacker.unpack_auth()
        try:
            meth() # Unpack args, call turn_around(), pack reply
        except (EOFError, RPCGarbageArgs):
            # Too few or too many arguments
            self.packer.reset()
            self.packer.pack_uint(xid)
            self.packer.pack_uint(msg_type.REPLY)
            self.packer.pack_uint(reply_stat.MSG_ACCEPTED)
            self.packer.pack_auth((auth_flavor.AUTH_NULL, make_auth_null()))
            self.packer.pack_uint(accept_stat.GARBAGE_ARGS)
        return self.packer.get_buf()

    def turn_around(self):
        try:
            self.unpacker.done()
        except RuntimeError:
            raise RPCGarbageArgs
        self.packer.pack_uint(accept_stat.SUCCESS)

    def handle_0(self): # Handle NULL message
        self.turn_around()

    def addpackers(self):
        # Override this to use derived classes from Packer/Unpacker
        self.packer = Packer()
        self.unpacker = Unpacker('')


class TCPServer(Server):

    def __init__(self, host, prog, vers, port):
        Server.__init__(self, host, prog, vers, port)
        self.connect()
    
    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.prot = IPPROTO_TCP
        self.sock.bind((self.host, self.port))

    def loop(self):
        self.sock.listen(0)
        while 1:
            self.session(self.sock.accept())

    def session(self, connection):
        sock, (host, port) = connection
        while 1:
            try:
                call = recvrecord(sock)
            except EOFError:
                break
            except socket.error:
                print('socket error:', sys.exc_info()[0])
                break
            reply = self.handle(call)
            if reply is not None:
                sendrecord(sock, reply)

    def forkingloop(self):
        # Like loop but uses forksession()
        self.sock.listen(0)
        while 1:
            self.forksession(self.sock.accept())

    def forksession(self, connection):
        # Like session but forks off a subprocess
        import os
        # Wait for deceased children
        try:
            while 1:
                pid, sts = os.waitpid(0, 1)
        except os.error:
            pass
        pid = None
        try:
            pid = os.fork()
            if pid: # Parent
                connection[0].close()
                return
            # Child
            self.session(connection)
        finally:
            # Make sure we don't fall through in the parent
            if pid == 0:
                os._exit(0)


class UDPServer(Server):

    def __init__(self, host, prog, vers, port):
        Server.__init__(self, host, prog, vers, port)
        self.connect()
    
    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.prot = IPPROTO_UDP
        self.sock.bind((self.host, self.port))
    
    def loop(self):
        while 1:
            self.session()

    def session(self):
        call, host_port = self.sock.recvfrom(8192)
        reply = self.handle(call)
        if reply is not None:
            self.sock.sendto(reply, host_port)



"""

    lantz.drivers.rigol.dg1022
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of Rigol DG1022 function generator with 2 channels. More or
    less based around on the Ag33522a driver from Berk Diler.

    Manual available from: https://www.rigolna.com/products/waveform-generators/dg1000/

    Author: Peter Mintun

    Date: 12/11/2017

"""


import numpy as np
import lantz
from lantz import Action, Feat, DictFeat, ureg
from collections import OrderedDict
from lantz.messagebased import MessageBasedDriver

class DG1022(MessageBasedDriver):

    DEFAULTS = {
        'COMMON': {
            'write_termination': '\n',
            'read_termination': '\n',
        }
    }

    CHANNELS = OrderedDict([(1, 1),
                           (2,2)])

    TOGGLE = OrderedDict([('on', 'ON'),
                          ('off', 'OFF')])

    WAVEFORMS = OrderedDict([('arbitrary', 'ARB'),
                             ('dc', 'DC'),
                             ('harmonic', 'HARM'),
                             ('noise', 'NOIS'),
                             ('pulse', 'PULS'),
                             ('ramp', 'RAMP'),
                             ('sine', 'SIN'),
                             ('square', 'SQU'),
                             ('triangle', 'TRI'),
                             ('user', 'USER')])

    @Feat(read_once=True)
    def idn(self):
        return self.query('*IDN?')

    @Action()
    def reset(self):
        return self.write('*RST')

    @Feat()
    def error(self):
        msg = self.query("SYST:ERR?")
        return msg.split(',') #error code, error message

    @DictFeat(keys = CHANNELS, units="Hz", limits=(1e-6, 25e6))
    def frequency(self,channel):
        """
        Returns the frequency of the specified channel, in Hertz.
        """
        return float(self.query('SOUR{}:FREQ?'.format(channel)))

    @frequency.setter
    def frequency(self,channel,value):
        """
        Sets the frequency of the specified channel, to value. Note that this
        is not smart enough to keep track of the different bandwidth constraints
        on different types of waveforms, so see the manual accordingly.
        """
        return self.write('SOUR{}:FREQ {:1.6f}'.format(channel,value))

    @DictFeat(keys=CHANNELS, values=WAVEFORMS)
    def function(self,channel):
        """
        Returns the function of the specified channel from the options
        enumerated in WAVEFORMS.
        """
        result = self.query('SOUR{}:APPL?'.format(channel))[1:-1]
        return result.split(',')[0]

    @function.setter
    def function(self,channel,value):
        """
        Returns the function of the specified channel to value (specified in
        WAVEFORMS).
        """
        return self.write('SOUR{}:APPL:{}'.format(channel, value))

    @DictFeat(keys = CHANNELS, values = TOGGLE)
    def output(self,channel):
        """
        Reads the output state of the specified channel.
        """
        return self.query('OUTP{}?'.format(channel))

    @output.setter
    def output(self,channel,val):
        """
        Sets the output state of the specified channel to val.
        """
        return self.write('OUTP{} {}'.format(channel,val))

    @DictFeat(keys=CHANNELS, units="V", limits=(-10.,10.))
    def voltage_low(self,channel):
        """
        Queries the low voltage level for the specified channel.
        """
        return float(self.query("SOUR{}:VOLT:LOW?".format(channel)))

    @voltage_low.setter
    def voltage_low(self,channel,value):
        """
        Sets the high voltage level for the specified channel.
        """
        return self.write("SOUR{}:VOLT:LOW {:1.6f}".format(channel, value))

    @DictFeat(keys=CHANNELS, units="V", limits=(-10.,10.))
    def voltage_high(self,channel):
        """
        Queries the high voltage level for the specified channel.
        """
        return float(self.query("SOUR{}:VOLT:HIGH?".format(channel)))

    @voltage_high.setter
    def voltage_high(self,channel,value):
        """
        Sets the high voltage level for the specified channel.
        """
        return self.write("SOUR{}:VOLT:HIGH {:1.6f}".format(channel, value))

    @DictFeat(keys = CHANNELS, units = "V", limits=(0., 20.))
    def voltage_amplitude(self,channel):
        """
        Queries the peak-to-peak voltage amplitude of the specified output
        channel.
        """
        return float(self.query("SOUR{}:VOLT?".format(channel)))

    @voltage_amplitude.setter
    def voltage_amplitude(self,channel,value):
        """
        Sets the peak-to-peak voltage amplitude of the specified output channel.
        """
        return self.write("SOUR{}:VOLT {:1.6f}".format(channel,value))

    @DictFeat(keys=CHANNELS, units="V", limits=(-10., 10.))
    def voltage_offset(self, channel):
        """
        Queries the offset voltage of the specified output channel.
        """
        return float(self.query("SOUR{}:VOLT:OFFS?".format(channel)))

    @voltage_offset.setter
    def voltage_offset(self, channel, value):
        """
        Sets the offset voltage of the specified output channel.
        """
        self.write("SOUR{}:VOLT:OFFS {:1.6f}".format(channel, value))


if __name__ == '__main__':


    # note: if you don't see your device, it may not work over USB 3.0?
    addr = 'USB0::0x1AB1::0x0642::DG1ZA192902819::INSTR'

    try:
        inst = DG1022(addr)
        inst.initialize()

        inst.reset()

        print('Identification:{}'.format(inst.idn))
        print('Error:{}'.format(inst.error))


    except:

        print('Could not find instrument, check connection/address!')

    # code to check various parameters from supported channels
    for channel in inst.CHANNELS.keys():

        inst.frequency[channel] = 1e-6
        print('Channel {} frequency: {}'.format(channel, inst.frequency[channel]))
        inst.frequency[channel] = 20e6
        print('Channel {} frequency: {}'.format(channel, inst.frequency[channel]))

        print('Channel {} function: {}'.format(channel, inst.function[channel]))
        inst.function[channel] = 'square'
        print('Channel {} function: {}'.format(channel, inst.function[channel]))

        inst.output[channel] = 'off'
        print('Channel {} output:{}'.format(channel, inst.output[channel]))
        inst.output[channel] = 'on'
        print('Channel {} output:{}'.format(channel, inst.output[channel]))
        inst.output[channel] = 'off'
        print('Channel {} output:{}'.format(channel, inst.output[channel]))

        print('Channel {} low voltage:{}'.format(channel, inst.voltage_low[channel]))
        inst.voltage_low[channel] = -1.0
        print('Channel {} low voltage:{}'.format(channel, inst.voltage_low[channel]))

        print('Channel {} high voltage:{}'.format(channel, inst.voltage_high[channel]))
        inst.voltage_high[channel] = 1.0
        print('Channel {} high voltage:{}'.format(channel, inst.voltage_high[channel]))

        print('Channel {} voltage amplitude:{}'.format(channel, inst.voltage_amplitude[channel]))
        print('Channel {} voltage offset:{}'.format(channel, inst.voltage_offset[channel]))

        inst.voltage_amplitude[channel] = 5.0
        inst.voltage_offset[channel] = 0.0

        print('Channel {} voltage amplitude:{}'.format(channel, inst.voltage_amplitude[channel]))
        print('Channel {} voltage offset:{}'.format(channel, inst.voltage_offset[channel]))
        print('Channel {} low voltage:{}'.format(channel, inst.voltage_low[channel]))
        print('Channel {} high voltage:{}'.format(channel, inst.voltage_high[channel]))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.pco.sensicam
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements Sensicam driver


    Implementation Notes
    --------------------

    The Sensicam operates using a COC = Command Operation Code

    ---

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from time import sleep
from collections import namedtuple

import ctypes as ct
import numpy as np

from lantz import Action, Feat
from lantz import errors
from lantz.foreign import LibraryDriver, RetTuple

class Status(object):
    READ_BUSY = 1
    BUFFER_EMPTY = 2
    COC_RUNNING = 4
    BUFFERS_FULL = 8

_ERRORS = {
        0: "Success",
        -1: "No camera connected",
        -2: "Timeout",
        -3: "Wrong parameter",
        -4: "cannot locate card",
        -5: "cannot allocate DMA buffer",
        -7: "DMA timeout",
        -8: "Invalid camera mode",
        -9: "No driver installed",
        -10: "No PCI bios found",
        -11: "Device is hold by another process",
        -12: "Error in reading or writing data to board",
        -13: "Wrong driver function",
        -20: "Load COC error",
        -21: "Too many values in COC",
        -22: "Temperatur out of range",
        -23: "Buffer allocate error",
        -24: "Read image error",
        -25: "Set/reset buffer flags is failed",
        -26: "Buffer is used",
        -27: "Call to a windows function is failed",
        -28: "DMA error",
        -29: "Cannot open file",
        -30: "Registry error",
        -31: "Open dialog error",
        -32: "Needs newer called vxd or dll",
        100: 'no image in PCI buffer',
        101: 'picture too dark',
        102: 'picture too bright',
        103: 'one or more values changed',
        104: 'buffer for builded string too short'}

#: Command Operation Code tuple
COC = namedtuple('COC', 'mode trigger roi binning table')


class Sensicam(LibraryDriver):
    """PCO Sensicam
    """

    LIBRARY_NAME = 'senntcam.dll'

    def __init__(self, board, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.board = board

    def _return_handler(self, func_name, ret_value):
        if ret_value < 0:
            raise errors.InstrumentError(_ERRORS[ret_value])
        elif (ret_value > 0 and
              func_name not in ('GET_COCTIME', 'GET_DELTIME', 'GET_BELTIME', 'GET_EXPTIME')):
            self.log_warning('While calling {}: {}', func_name, _ERRORS[ret_value])

        return ret_value

    def initialize(self):
        self.lib.SET_BOARD(_boardnr)
        err = self.lib.SET_INIT(1)
        self.stop_coc()

    def finalize(self):
        self.stop_coc()
        self.lib.SET_INIT(0)

    @Feat()
    def status(self):
        # Get camera type, temperature of electronics and temp. of CCD
        cam_type = ct.pointer(ct.c_int())
        temp_ele = ct.pointer(ct.c_int())
        temp_ccd = ct.pointer(ct.c_int())
        err = self.lib.GET_STATUS(cam_type, temp_ele, temp_ccd)
        #return "{}".format(_ERRORS[err])
        return cam_type, temp_ele, temp_ccd

    @Feat()
    def mode(self):
        """Imaging mode as a 3-element tuple (type, gain and submode).
        """
        _mode = self.coc.mode
        return _mode & 0xFF, (_mode >> 8) & 0xFF, (_mode >> 16) & 0xFF

    @mode.setter
    def mode(self, typ_gain_submode):
        typ, gain, submode = typ_gain_submode
        _mode = (typ & 0xFF) | ((gain & 0xFF) << 8) | ((submode & 0xFF) << 16)
        self.coc = self.recall('coc')._replace(mode=_mode)

    @Feat()
    def trigger(self):
        """Triger mode.
        """
        return self.coc.trigger

    @trigger.setter
    def trigger(self, value):
        self.coc = self.recall('coc')._replace(trigger=value)

    @Feat()
    def roi(self):
        """Region of interest in pixels as a 4-element tuple (x1, x2, y1, y2).
        """
        return self.coc.roi

    @roi.setter
    def roi(self, roi):
        roix1, roix2, roiy1, roiy2 = roi
        self.coc = self.recall('coc')._replace(roi=(roix1, roix2, roiy1, roiy2))

    @Feat()
    def binning(self):
        """Binning in pixels as a 2-element tuple (horizontal, vertical).
        """
        return self.coc.binning

    @binning.setter
    def binning(self, hbin_vbin):
        hbin, vbin = hbin_vbin
        self.coc = self.recall('coc')._replace('binning', (hbin, vbin))

    @Feat(units='ms')
    def exposure_time(self):
        """Exposure time.
        """
        return float(self.coc.table.split(',')[1])

    @exposure_time.setter
    def exposure_time(self, value):
        self.table = '0,{},-1,-1'.format(value)

    @Feat()
    def table(self):
        """COC table
        """
        return self.coc.table

    @Action()
    def stop_coc(self):
        # Stop camera
        self.lib.STOP_COC(0)

    @Action()
    def run_coc(self, runmode = 4):
         # 0 continuous - 4 single
        self.lib.RUN_COC(runmode)
        
    @Feat()
    def coc(self):
        """Command Operation Code
        """

        mode = ct.pointer(ct.c_int())
        trig = ct.pointer(ct.c_int())
        roix1 = ct.pointer(ct.c_int())
        roix2 = ct.pointer(ct.c_int())
        roiy1 = ct.pointer(ct.c_int())
        roiy2 = ct.pointer(ct.c_int())
        hbin = ct.pointer(ct.c_int())
        vbin = ct.pointer(ct.c_int())
        table = ct.create_string_buffer(40)
        self.lib.GET_SETTINGS(mode, trig, roix1, roix2, roiy1, roiy2,
                               hbin, vbin, table) #Todo: Try RetTuple

        return COC(mode[0], trig[0], (roix1[0], roix2[0], roiy1[0], roiy2[0]),
            (hbin[0], vbin[0]), table[0].decode('ascii'))

    @coc.setter
    def coc(self, value):
        newcoc = (value.mode, value.trigger) + value.roi + \
                  value.binning + (value.table, )
        self.lib.SET_COC(*newcoc)
        
    @Feat()
    def image_status(self):
        """Image status
        """
        stat = ct.pointer(ct.c_int())
        self.lib.GET_IMAGE_STATUS(stat)
        return stat

    @Feat()
    def image_size(self):
        """Image size in pixels (width, height).
        """
        width = ct.pointer(ct.c_int())
        height = ct.pointer(ct.c_int())
        self.lib.GET_IMAGE_SIZE(width, height)

        return width[0], height[0]

    @Feat(units='microseconds') #TODO: check units
    def coc_time(self):
        return self.lib.GET_COCTIME()

    @Feat(units='microseconds')
    def delay_time(self):
        return self.lib.GET_DELTIME()

    @Feat(units='microseconds')
    def exp_time(self):
        return self.lib.GET_EXPTIME()

    @Feat(units='microseconds')
    def bel_time(self):
        return self.lib.GET_BELTIME()

    @Action(units='ms')
    def expose(self, exposure = 1):
        """Expose.

        :param exposure: exposure time.

        """
        self.exposure_time = exposure
        delay = self.coc_time()
        self.run_coc()
        sleep(delay / 1000)
        while not (self.image_status & Status.COC_RUNNING):
            sleep(delay / 1000 * .1)

    @Action()
    def read_out(self):
        """Readout image from the CCD.

        :rtype: NumPy array
        """
        width, height = self.image_size
        imagearray = np.zeros((width,height), dtype=np.dtype(np.ushort))
        image = np.ascontiguousarray(imagearray)
        ptrimage = image.ctypes.data_as(ct.POINTER(ct.c_ushort))
        self.lib.READ_IMAGE_12BIT(0, width, height, ptrimage)
        return image


    @Action(units='ms')
    def take_image(self, exposure=1):
        """Take image.

        :param exposure: exposure time.
        :rtype: NumPy array
        """
        self.expose(exposure)
        return self.read_out()


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-b', '--board', type=int, default='0',
                        help='Board number')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with Sensicam(args.board) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            import matplotlib.pyplot as plt

            image = inst.take_image(10)

            print(np.min(image), np.max(image), np.mean(image))

            plt.imshow(image, cmap='gray')
            plt.show()



# -*- coding: utf-8 -*-
"""
    lantz.drivers.rigol.ds1052e
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control an oscilloscope.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
    
    Source: DS1052e manual
"""


from lantz import MessageBasedDriver


class DS1052e(MessageBasedDriver):
    pass



if __name__ == '__main__':
    with DS1052e('USB0::0x1AB1::0x0588::DS1K00005888::INSTR') as inst:
        print(inst.query('*IDN?'))



"""
    lantz.drivers.pi.piezo
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Implements the drivers to control pi E-709 piezo motion controller
    via USB or via serial. It uses the PI General Command Set.

    For USB, one first have to install the driver from PI.
    :copyright: 2018, see AUTHORS for more details.
    :license: GPL, see LICENSE for more details.

    Source: Instruction Manual (PI)
"""

from lantz.feat import Feat
from lantz.action import Action
from lantz.messagebased import MessageBasedDriver
# from pyvisa import constants
from lantz import Q_, ureg
# from lantz.processors import convert_to
import time
import numpy as np
# import copy
from collections import OrderedDict

if __name__=='__main__':
    from error_codes import error_codes
else:
    from .error_codes import error_codes


def parse_line(line):
    line = line.strip()  # remove whitespace at end
    return line.split('=')


def parse_multi(message):
    '''Return an ordered dictionary containing the returned parameters'''
    assert isinstance(message, list)
    return OrderedDict([parse_line(line) for line in message])


class Piezo(MessageBasedDriver):
    """ PI piezo motion controller. It assumes all axes to have units um

    Important:
    before operating, ensure that the notch filters are set to the right 
    frequencies. These frequencies depend on the load on the piezo, and can 
    be determined in NanoCapture.
    
    Example:
    import numpy as np
    import lantz
    import visa
    import lantz.drivers.pi.piezo as pi
    rm = visa.ResourceManager('@py')
    lantz.messagebased._resource_manager = rm
    ureg = lantz.ureg
    try:
        lantzlog
    except NameError:
        #lantzlog = lantz.log.log_to_screen(level=lantz.log.INFO)
        lantzlog = lantz.log.log_to_screen(level=lantz.log.ERROR)

    import warnings
    warnings.filterwarnings(action='ignore')

    print('Available devices:', lantz.messagebased._resource_manager.list_resources())
    print('Connecting to: ASRL/dev/ttyUSB0::INSTR')
    stage = pi.Piezo('ASRL/dev/ttyUSB0::INSTR', axis='X',
        sleeptime_after_move=10*ureg.ms)
    stage.initialize()
    idn = stage.idn

    stage.servo = True
    stage.velocity = 10*ureg.um/ureg.ms

    stage.position = 0*ureg.um
    print('stage position measured: ', stage.position)

    steps = 10
    stepsize = 100.0*ureg.nm
    for n in range(steps):
        stage.position = n*stepsize
        print('stage position measured: ', stage.position)

    print('Measuring step response')

    # For jupyter notebooks
    %matplotlib inline

    stage.servo = True
    stage.position = 0
    import time
    time.sleep(1)

    stepsize = 10*lantz.ureg.um
    timepoints = 100

    timepos = stage.measure_step_response(stepsize, timepoints)

    lines = stage.plot_step_response(timepos)

    stage.finalize()  # nicely close stage and turn of servo
    
    
    """

    DEFAULTS = {'COMMON': {'write_termination': '\n',
                           'read_termination': '\n',
                           'baud_rate': 57600,
                           'timeout': 20}, }

    def __init__(self, *args, **kwargs):
        self.sleeptime_after_move = kwargs.pop('sleeptime_after_move', 0*ureg.ms)
        self.axis = kwargs.pop('axis', 'X')
        super().__init__(*args, **kwargs)

    def initialize(self):
        super().initialize()

    def finalize(self):
        """ Disconnects stage """
        self.stop()
        super().finalize()

    def query_multiple(self, query):
        """Read a multi line response"""
        self.write(query)
        loop = True
        ans = []
        while loop:
            ret = self.read()
            if ret != '':
                ans.append(ret)
            else:
                return ans

    def parse_multiaxis(self, message, axis=None):
        "Parse a multi-axis message, return only value for self.axis"
        if axis is None:
            axis = self.axis
        message_parts = message.split(' ')
        try:
            message_axis = [part for part in message_parts if part[0] == axis][0]
        except IndexError:
            self.log_error('Axis {} not found in returned message: {}'.format(axis, message))
        values = message_axis.split('=')
        return values[-1]

    @Feat()
    def errors(self):
        error = int(self.query('ERR?'))
        if error != 0:
            self.log_error('Stage error code {}: {}'.format(error, error_codes[error]))
        return error

    @Feat()
    def idn(self):
        idn = self.query("*IDN?")
        self.errors
        return idn

    @Action()
    def stop(self):
        '''Stop all motions'''
        self.servo = False
        self.write('STP')
        self.errors

    @Feat(values={True: '1', False: '0'}, read_once=True)
    def servo(self):
        ''' Set the stage control in open- or closed-loop (state = False or True)'''
        return self.parse_multiaxis(self.query('SVO?'))

    @servo.setter
    def servo(self, state):
        self.write('SVO {} {}'.format(self.axis, state) )
        return self.errors

    @Feat(units='um/s')
    def velocity(self):
        ''' Set the stage velocity (closed-loop only)'''
        return self.parse_multiaxis(self.query('VEL?'))

    @velocity.setter
    def velocity(self, velocity):
        self.write('VEL {} {}'.format(self.axis, velocity))
        return self.errors

    @Feat(units='um')
    def position(self):
        ''' Move to an absolute position the stage (closed-loop only)'''
        return self.parse_multiaxis(self.query('POS?'))

    @position.setter
    def position(self, position):
        return self.move_to(position, self.sleeptime_after_move)

    @Action(units=('um','ms'))
    def move_to(self, position, timeout=None):
        ''' Move to an absolute position the stage (closed-loop only)'''
        self.write('MOV {} {}'.format(self.axis, position))
        time.sleep(timeout * 1e-3) # Give the stage time to move! (in seconds!)
        return self.errors

    @Feat(units='um')
    def read_stage_position(self, nr_avg = 1):
        ''' Read the current position from the stage'''
        positions = [self.position.magnitude for n in range(nr_avg)]
        return np.mean(positions)

    @Action(units=('um',None))
    def measure_step_response(self, stepsize, points):
        '''Python implementation to measure a step response of size  stepsize.
        Measure number of points as fast as possible.

        Servo should be on

        Note: a higher temporal resolution can be aquired by the data-recorder
        in the driver

        Example:
        stage.servo = True
        stage.position = 0
        time.sleep(1)

        stepsize = 10*lantz.ureg.um
        timepoints = 100

        timepos = stage.measure_step_response(stepsize, timepoints)

        lines = stage.plot_step_response(timepos)
        '''
        if not self.servo:
            self.log.error('Servo should be on')
            return
        import time
        self.move_to(self.position + stepsize*ureg.um, 0)
        timepos = np.array([[time.time(), self.position.magnitude] for i in range(points)])
        timepos[:,0] -= timepos[0,0] #correct time offset
        timepos[:,0] *= 1000 # make it milliseconds

        return timepos

    def plot_step_response(self, timepos):
        '''Helper function to visualize a step respons'''
        import matplotlib.pyplot as plt
        lines = plt.plot(*timepos.T)
        plt.xlabel('Time [ms]')
        plt.ylabel('Position [um]')
        return lines

if __name__=='__main__':
    # before operating, ensure that the notch filters are set to the right 
    # frequencies. These frequencies depend on the load on the piezo, and can 
    # be determined in NanoCapture.
    from lantz.ui.app import start_test_app

    import lantz
    import visa
    import lantz.drivers.pi.piezo as pi
    rm = visa.ResourceManager('@py')
    lantz.messagebased._resource_manager = rm
    try:
        lantzlog
    except NameError:
        lantzlog = lantz.log.log_to_screen(level=lantz.log.INFO)
    with Piezo('ASRL/dev/ttyUSB0::INSTR') as inst:
        start_test_app(inst)




from lantz import Feat, Action
from lantz.messagebased import MessageBasedDriver
from time import sleep
from pyvisa import constants


class Newport69907(MessageBasedDriver):
    """
    This class implements serial control of a Newport69907 arc lamp power
    supply.
    """
    DEFAULTS = {'ASRL': {'write_termination': '\r',
                         'read_termination': '\r',
                         'baud_rate': 9600,
                         'data_bits': 8,
                         'parity': constants.Parity.none,
                         'stop_bits': constants.StopBits.one,
                         'encoding': 'latin-1',
                         'timeout': 10000}}

    max_lamp_amps = 12  # amps
    max_lamp_power = 100  # watts

    def parse_status_register(self, status_register):
        """
        Parses the status registry codes bit by bit into meaningful statement.
        These error bits correspond to whether or not the corresponding LEDs
        on the front panel are illuminated.

        Bit 7 - Lamp on
        Bit 6 - Ext
        Bit 5 - Power/Current mode
        Bit 4 - Cal mode
        Bit 3 - Fault
        Bit 2 - Communication
        Bit 1 - Limit
        Bit 0 - Interlock
        """
        status_code = int(('0x' + status_register.replace('STB', '')), 16)
        status = []
        # print('Status code:{}'.format(status_code))
        if status_code > 128:
            status.append('Lamp on')
            status_code -= 128
        if status_code > 64:
            status.append('Ext')
            status_code -= 64
        if status_code > 32:
            status.append('Power Mode')
            status_code -= 32
        else:
            status.append('Current Mode')
        if status_code > 16:
            status.append('Cal Mode')
            status_code -= 16
        if status_code > 8:
            status.append('Fault')
            status_code -= 8
        if status_code > 4:
            status.append('Comm')
            status_code -= 4
        if status_code > 2:
            status.append('Limit')
            status_code -= 2
        if status_code == 1:
            status.append('Interlock')
        return status

    def parse_error_register(self, error_register):
        """
        Parses the status registry codes bit by bit into meaningful statement.

        Bit 7 - Power on
        Bit 6 - User request
        Bit 5 - Command error
        Bit 4 - Execution error
        Bit 3 - Device dependent error
        Bit 2 - Query error
        Bit 1 - Request control
        Bit 0 - Operation complete
        """
        # convert hex string into integer
        err_code = int(('0x' + error_register.replace('ESR', '')), 16)
        errors = []
        # print('error code:{}'.format(err_code))
        if err_code > 128:
            errors.append('Power on')
            err_code -= 128
        if err_code > 64:
            errors.append('User request')
            err_code -= 64
        if err_code > 32:
            errors.append('Command error')
            err_code -= 32
        if err_code > 16:
            errors.append('Execution error')
            err_code -= 16
        if err_code > 8:
            errors.append('Device dependent error.')
            err_code -= 8
        if err_code > 4:
            errors.append('Query error.')
            err_code -= 4
        if err_code > 2:
            errors.append('Request control')
            err_code -= 2
        # if err_code == 1:
            # no error
        return errors

    @Feat()
    def idn(self):
        """
        Return power supply model number
        """
        return self.query('IDN?')

    @Feat()
    def status(self):
        """
        Returns status information of the instrument.
        """
        status = self.query('STB?')
        return self.parse_status_register(status)

    @Feat()
    def error_status(self):
        """
        Returns the instrument error status.
        """
        error_register = self.query('ESR?')
        return self.parse_error_register(error_register)

    @Feat()
    def amps(self):
        """
        Returns output amperage displayed on the front panel.
        """
        return float(self.query('AMPS?'))

    @Feat()
    def volts(self):
        """
        Return output voltage displayed on front panel.
        """
        return float(self.query('VOLTS?'))

    @Feat()
    def watts(self):
        """
        Returns output wattage displayed on front panel.
        """
        return int(self.query('WATTS?'))

    @Feat()
    def lamp_hrs(self):
        """
        Returns the number of hours on the current lamp.
        """
        return self.query('LAMP HRS?')

    @Feat()
    def amp_preset(self):
        """
        Returns the lamp amperage preset value.
        """
        return float(self.query('A-PRESET?'))

    @amp_preset.setter
    def amp_preset(self, preset_val):
        """
        Sets the lamp amperage preset value to preset_val.
        """
        error_status = self.query('A-PRESET={0:.1f}'.format(preset_val))
        return self.parse_error_register(error_status)

    @Feat()
    def power_preset(self):
        """
        Returns the lamp power preset value (in watts).
        """
        return int(self.query('P-PRESET?'))

    @power_preset.setter
    def power_preset(self, power_preset):
        """
        Sets the lamp powr preset value (in watts).
        """
        error_status = self.query('P-PRESET={0:04d}'.format(int(power_preset)))
        return self.parse_error_register(error_status)

    @Feat(limits=(0, max_lamp_amps))
    def amp_lim(self):
        """
        Return current lamp amperage limit.
        """
        return float(self.query('A-LIM?'))

    @amp_lim.setter
    def amp_lim(self, max_amps):
        """
        Sets current amperage limit to max_amps
        """
        error_register = self.query('A-LIM={0:.1f}'.format(max_amps))
        return self.parse_error_register(error_register)

    @Feat(limits=(0, max_lamp_power))
    def power_lim(self):
        """
        Return current power limit.
        """
        return int(self.query('P-LIM?'))

    @power_lim.setter
    def power_lim(self, max_power):
        """
        Sets the power limit of the lamp to max_power.
        """
        error_register = self.query('P-LIM={0:04d}'.format(int(max_power)))
        return self.parse_error_register(error_register)

    @Action()
    def start_lamp(self):
        """
        Turns on the arc lamp.
        """
        error_register = self.query('START')
        return self.parse_error_register(error_register)

    @Action()
    def stop_lamp(self):
        """
        Turns off the arc lamp.
        """
        error_register = self.query('STOP')
        return self.parse_error_register(error_register)

    @Action()
    def reset(self):
        """
        Returns the arc lamp to factory default settings.
        """
        error_register = self.query('RST')
        return self.parse_error_register(error_register)


if __name__ == '__main__':
    with Newport69907('ASRL9::INSTR') as inst:
        print('== Instrument Information ==')
        print('Serial number:{}'.format(inst.idn))
        print('Status:{}'.format(inst.status))
        print('Errors:{}'.format(inst.error_status))

        print('== Current lamp status ==')
        print('Amps:{}A'.format(inst.amps))
        print('Volts:{}V'.format(inst.volts))
        print('Watts:{}W'.format(inst.watts))
        print('Lamp hours:{}hrs'.format(inst.lamp_hrs))

        print('== Lamp Settings ==')
        print('Current preset:{}A'.format(inst.amp_preset))
        print('Power preset:{}W'.format(inst.power_preset))
        print('Current limit:{}A'.format(inst.amp_lim))
        print('Power limit:{}W'.format(inst.power_lim))

        print('== Lamp Control ==')
        print('Starting lamp...')
        inst.start_lamp()
        sleep(5)
        print('Amps:{}A'.format(inst.amps))
        print('Volts:{}V'.format(inst.volts))
        print('Watts:{}W'.format(inst.watts))
        print('Stopping lamp...')
        inst.stop_lamp()

        print('== Changing Lamp Settings ==')
        inst.amp_lim = 12.0
        inst.amp_preset = 7.5
        inst.power_lim = 98
        inst.power_preset = 95

        print('Current preset:{}A'.format(inst.amp_preset))
        print('Power preset:{}W'.format(inst.power_preset))
        print('Current limit:{}A'.format(inst.amp_lim))
        print('Power limit:{}W'.format(inst.power_lim))

        inst.amp_lim = 10.0
        inst.amp_preset = 5.5
        inst.power_lim = 80
        inst.power_preset = 75

        print('Current preset:{}A'.format(inst.amp_preset))
        print('Power preset:{}W'.format(inst.power_preset))
        print('Current limit:{}A'.format(inst.amp_lim))
        print('Power limit:{}W'.format(inst.power_lim))

        print('Starting lamp...')
        inst.start_lamp()
        sleep(5)
        print('Amps:{}A'.format(inst.amps))
        print('Volts:{}V'.format(inst.volts))
        print('Watts:{}W'.format(inst.watts))
        print('Stopping lamp...')
        inst.stop_lamp()
        sleep(1)

        inst.amp_lim = 12.0
        inst.amp_preset = 7.5
        inst.power_lim = 98
        inst.power_preset = 95

        print('Current preset:{}A'.format(inst.amp_preset))
        print('Power preset:{}W'.format(inst.power_preset))
        print('Current limit:{}A'.format(inst.amp_lim))
        print('Power limit:{}W'.format(inst.power_lim))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.labjack._internal
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Python drivers provided by LabJack.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""



# -*- coding: utf-8 -*-
"""
    AG-UC8.py

    :brief Implements the drivers for the agilis controller

    :author AlexBourassa
"""

from lantz import Feat, DictFeat, Action
from lantz.errors import InstrumentError
from lantz.messagebased import MessageBasedDriver
import re

import pyvisa.resources.serial as _s

speed_dict = {'stop':0, 'lowest':1, 'low':2, 'high':3, 'highest':4}
inv_speed_dict = {v: k for k, v in speed_dict.items()}

jog_modes = {'no move': '0','1700 step/s' : '3', '666 step/s' : '4',  '100 step/s' : '2',  '5 step/s': '1',
                            '-1700 step/s': '-3','-666 step/s': '-4', '-100 step/s': '-2', '-5 step/s': '-1'}

class AG_UC8(MessageBasedDriver):
    """
    Implements the drivers for the agilis controller
    """
    
    DEFAULTS = {'ASRL': {'write_termination': '\r\n',
                         'read_termination': '\r\n',
                         'baud_rate': 921600,
                         'data_bits' : 8,
                         'stop_bits' : _s.constants.StopBits.one,
                         'timeout' : 20,
                         'parity' : _s.constants.Parity.none}}

    speed_list = [[0,0]]*4 # Instance variable that defines the speeds which will be used for all the move calls

    def __init__(self, resource_name, name='AG-UC8',**kwargs):
        MessageBasedDriver.__init__(self, resource_name=resource_name, name=name, **kwargs)

    def query(self, cmd, send_args=(None, None), recv_args=(None, None), filter=r'.*'):
        """Sends a query using the base class method, then parses the response according to the regex filter and
           returns group 1 (or group zero if no group in the filter)
        """
        ans = MessageBasedDriver.query(self, command=cmd,send_args=send_args, recv_args=recv_args)
        match = re.search(filter, ans)
        if match is None            : raise Exception("Filter '" +filter+"' didn't match ans '" + ans +"'")
        elif len(match.groups())==0 : return match.group(0)
        else                        : return match.group(1)



    @DictFeat(keys = [1,2,3,4], values=speed_dict)
    def speed1(self, channel):
        """Speed of axis 1 for a given channel
        """
        return self.speed_list[channel-1][0]

    @speed1.setter
    def speed1(self, channel, speed):
        self.speed_list[channel][0] = speed

    @DictFeat(keys = [1,2,3,4], values=speed_dict)
    def speed2(self, channel):
        """Speed of axis 2 for a given channel
        """
        return self.speed_list[channel-1][1]

    @speed2.setter
    def speed2(self, channel, speed):
        self.speed_list[channel][1] = speed

    @Feat()
    def limit_status(self):
        """Tell limit status
        """
        return self.query('PH?')

    @Action()
    def jog(self, axis=1, reverse=False):
        """Move the current channel specified axis at the specified speed
        """
        speed = self.speed_list[int(self.channel)-1][axis-1]
        if reverse and speed!=0: cmd = '{}JA-{}'
        else      : cmd = '{}JA{}'
        print(self.query(cmd.format(axis,speed)))

    @Feat(values = {str(i):str(i) for i in [1,2,3,4]})
    def channel(self):
        return self.query('CC?', filter='CC([1,4])')



    @channel.setter
    def channel(self, channel):
        self.write('CC'+str(channel))



    def get_error_status(self,read_only=False):
        """Get error of previous command
             0  No error
            -1  Unknown command
            -2  Axis out of range (must be 1 or 2, or must not be specified)
            -3  Wrong format for parameter nn (or must not be specified)
            -4  Parameter nn out of range
            -5  Not allowed in local mode
            -6  Not allowed in current state
        """
        if read_only: return self.read()
        else        : return self.query('TE?')



    @Action()
    def stop_all(self):
        """Throw all the stop command in the book!
            Hopefully one of them works...
        """
        for i in [1,2]: self.move_to_limit[i] = 'no move'
        for i in [1,2]: self.jog = 'stop'
        # for stop_cmd in ['1MV0','2MV0','1JA0','2JA0']:
        #     self.write(stop_cmd)


    def check_connect(self):
        pass


# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.andor.neo
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements a high level driver for the Andor Neo CMOS Camera


    Sources::

        - Andor Neo Manual

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


import ctypes as ct

import numpy as np

from lantz import Feat, Action, Q_
from lantz.foreign import RetStr, RetTuple

from .andor import Andor

class Neo(Andor):
    """Neo Andor CMOS Camera
    """

    def initialize(self):
        super().initialize()
        self.flush()
        self.fan_speed = 1
        self.width ,self.height = self.sensor_size
        self.length = self.width * self.height
        self.clock_rate = 100
        self.pixel_encoding = 32
        self.imagesizebytes = self.getint("ImageSizeBytes")
        self.userbuffer = ct.create_string_buffer(' ' * self.imagesizebytes)

    @Feat(None, values={32: 'Mono32', 64: 'Mono64'})
    def pixel_encoding(self, value):
        """Pixel encoding.
        """
        self.setenumstring("PixelEncoding", value)

    @Feat()
    def sensor_size(self):
        width = self.getint("SensorWidth")
        height = self.getint("SensorHeight")        
        return width, height

    @Feat(None, values={100: '100 MHz', 200: '200 MHz', 280: '280 MHz'})
    def clock_rate(self, value):
        """Pixel clock rate
        """
        self.setenumstring("PixelReadoutRate", value)

    @Feat(None)
    def fan_peed(self, value = 1):
        """Fan speed.
        """
        self.setenumerated("FanSpeed", value)

    @Feat()
    def sensor_temp(self):
        """Sensor temperature.
        """
        return self.getfloat("SensorTemperature")

    @Feat()
    def exposure_time(self):
        """Get exposure time.
        """
        return self.getfloat("ExposureTime")

    @exposure_time.setter
    def exposure_time(self, exposure):
        self.setfloat("ExposureTime", exposure)

    @Feat(None)
    def roi(self, width_height_top_left):
        """Set region of interest
        """
        width, height, top, left = width_height_top_left
        self.setint("AOIWidth", width)
        self.setint("AOILeft", left)
        self.setint("AOIHeight", height)
        self.setint("AOITop", top)

    @Action()
    def take_image(self):
        """Image acquisition.
        """
        self.queuebuffer(self.userbuffer, self.imagesizebytes)
        self.command("AcquisitionStart")
        self.waitbuffer(*RetStr(1))
        self.command("AcquisitionStop")
        self.flush()
        image = np.fromstring(self.userbuffer, dtype=np.uint32, count=self.length)
        image.shape = (self.height, self.width)
        return image

    @Action()
    def take_image(self, numbuff, numframes):
        """Image acquisition with circular buffer.
        """
        imagesizebytes = self.getint("ImageSizeBytes")
        userbuffer = []
        for i in range(numbuff):
            userbuffer.append(ct.create_string_buffer(' ' * imagesizebytes))
        self.queuebuffer(userbuffer, imagesizebytes)
        self.command("AcquisitionStart")
        for i in range(numbuff):
            self.waitbuffer(*RetStr(1))
            self.queuebuffer(userbuffer[i], imagesizebytes)
        self.command("AcquisitionStop")
        self.flush()
        image = np.fromstring(userbuffer[0], dtype=np.uint32, count=self.length)
        image.shape = (self.height, self.width)
        return image



# -*- coding: utf-8 -*-
"""
    lantz.drivers.stanford.sg396
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of SG396 signal generator

    Author: Kevin Miao & Berk Diler
    Date: 12/15/2015 & 8/21/17
"""

import numpy as np
from lantz import Action, Feat, DictFeat, ureg
from lantz.messagebased import MessageBasedDriver
from collections import OrderedDict

class SG396(MessageBasedDriver):

        DEFAULTS = {
            'COMMON': {
                'write_termination': '\r\n',
                'read_termination': '\r\n',
            }
        }

        MODULATION_TYPE = OrderedDict([
            ('AM', 0),
            ('FM', 1),
            ('Phase',2),
            ('Sweep',3),
            ('Pulse',4),
            ('Blank',5),
            ('QAM',7),
            ('CPM',8),
            ('VSB',9)
        ])

        MODULATION_FUNCTION = OrderedDict([
            ('sine', 0),
            ('ramp', 1),
            ('triangle', 2),
            ('square', 3),
            ('noise', 4),
            ('external', 5)
        ])

        # Signal synthesis commands

        @Feat
        def lf_amplitude(self):
            """
            low frequency amplitude (BNC output)
            """
            return float(self.query('AMPL?'))

        @lf_amplitude.setter
        def lf_amplitude(self, value):
            self.write('AMPL{:.2f}'.format(value))

        @Feat
        def rf_amplitude(self):
            """
            RF amplitude (Type N output)
            """
            return float(self.query('AMPR?'))

        @rf_amplitude.setter
        def rf_amplitude(self, value):
            self.write('AMPR{:.2f}'.format(value))

        @Feat(values={True: '1', False: '0'})
        def lf_toggle(self):
            """
            low frequency output state
            """
            return self.query('ENBL?')

        @lf_toggle.setter
        def lf_toggle(self, value):
            self.write('ENBL{:s}'.format(value))

        @Feat(values={True: '1', False: '0'})
        def rf_toggle(self):
            """
            RF output state
            """
            return self.query('ENBR?')

        @rf_toggle.setter
        def rf_toggle(self, value):
            self.write('ENBR{:s}'.format(value))

        @Feat(units='Hz')
        def frequency(self):
            """
            signal frequency
            """
            return self.query('FREQ?')

        @frequency.setter
        def frequency(self, value):
            self.write('FREQ{:.2f}'.format(value))

        @Feat()
        def lf_offset(self):
            """
            low frequency offset voltage
            """
            return self.query('OFSL?')

        @lf_offset.setter
        def lf_offset(self, value):
            self.write('OFSL{:.2f}'.format(value))

        @Feat(units='degrees')
        def phase(self):
            """
            carrier phase
            """
            return self.query('PHAS?')

        @phase.setter
        def phase(self, value):
            self.write('PHAS{:.2f}'.format(value))

        @Action()
        def rel_phase(self):
            """
            sets carrier phase to 0 degrees
            """
            self.write('RPHS')

        @Feat(values={True: 1, False: 0})
        def mod_toggle(self):
            """
            Modulation State
            """
            return int(self.query('MODL?'))

        @mod_toggle.setter
        def mod_toggle(self, value):
            self.write('MODL {}'.format(value))

        @Feat(values=MODULATION_TYPE)
        def mod_type(self):
            """
            Modulation State
            """
            return int(self.query('TYPE?'))

        @mod_type.setter
        def mod_type(self, value):
            self.write('TYPE {}'.format(value))

        @Feat(values=MODULATION_FUNCTION)
        def mod_function(self):
            """
            Modulation Function
            """
            return int(self.query('MFNC?'))

        @mod_function.setter
        def mod_function(self, value):
            self.write('MFNC {}'.format(value))

        @Feat(units="Hz", limits=(0.1, 100.e3))
        def mod_rate(self):
            """
            Modulation Rate
            """
            return float(self.query('RATE?'))

        @mod_rate.setter
        def mod_rate(self, val):
            self.write('RATE {}'.format(val))

        @Feat(limits=(0., 100.))
        def AM_mod_depth(self):
            """
            AM Modulation Depth
            """
            return float(self.query('ADEP?'))

        @AM_mod_depth.setter
        def AM_mod_depth(self, val):
            self.write('ADEP {}'.format(val))

        @Feat(units="Hz", limits=(0.1, 8.e6))
        def FM_mod_dev(self):
            """
            FM Modulation Deviation
            """
            return float(self.query("FDEV?"))

        @FM_mod_dev.setter
        def FM_mod_setter(self, val):
            self.write('FDEV {}'.format(val))





# -*- coding: utf-8 -*-
"""
    lantz.drivers.legacy.textual
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements a Mixin class for message based instruments,

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import time
from lantz.errors import LantzTimeoutError
from lantz.processors import ParseProcessor


class TextualMixin(object):
    """Mixin class for classes that communicate with instruments
    exchanging text messages.

    Ideally, transport classes should provide receive methods
    that support:
    1. query the number of available bytes
    2. read chunks of bytes (with and without timeout)
    3. read until certain character is found (with and without timeout)

    Most transport layers support 1 and 2 but not all support 3 (or only
    for a defined set of characters) and TextualMixin provides fallback
    a method.
    """

    #: Encoding to transform string to bytes and back as defined in
    #: http://docs.python.org/py3k/library/codecs.html#standard-encodings
    ENCODING = 'ascii'
    #: Termination characters for receiving data, if not given RECV_CHUNK
    #: number of bytes will be read.
    RECV_TERMINATION = ''
    #: Termination characters for sending data
    SEND_TERMINATION = ''
    #: Timeout in seconds of the complete read operation.
    TIMEOUT = 1
    #: Parsers
    PARSERS = {}
    #: Size in bytes of the receive chunk (-1 means all bytes in buffer)
    RECV_CHUNK = 1

    #: String containing the part of the message after RECV_TERMINATION
    #: Used in software based finding of termination character when
    #: RECV_CHUNK > 1
    _received = ''

    def raw_recv(self, size):
        """Receive raw bytes from the instrument. No encoding or termination
        character should be applied.

        This method must be implemented by base classes.

        :param size: number of bytes to receive.
        :return: received bytes, eom
        :rtype: bytes, bool
        """
        raise NotImplemented

    def raw_send(self, data):
        """Send raw bytes to the instrument. No encoding or termination
        character should be applied.

        This method must be implemented by base classes.

        :param data: bytes to be sent to the instrument.
        :param data: bytes.
        """
        raise NotImplemented

    def send(self, command, termination=None, encoding=None):
        """Send command to the instrument.

        :param command: command to be sent to the instrument.
        :type command: string.

        :param termination: termination character to override class defined
                            default.
        :param encoding: encoding to transform string to bytes to override class
                         defined default.

        :return: number of bytes sent.

        """
        if termination is None:
            termination = self.SEND_TERMINATION
        if encoding is None:
            encoding = self.ENCODING

        message = bytes(command + termination, encoding)
        self.log_debug('Sending {}', message)
        return self.raw_send(message)

    def recv(self, termination=None, encoding=None, recv_chunk=None):
        """Receive string from instrument.

        :param termination: termination character (overrides class default)
        :type termination: str
        :param encoding: encoding to transform bytes to string (overrides class default)
        :param recv_chunk: number of bytes to receive (overrides class default)
        :return: string encoded from received bytes
        """

        termination = termination or self.RECV_TERMINATION
        encoding = encoding or self.ENCODING
        recv_chunk = recv_chunk or self.RECV_CHUNK

        if not termination:
            return str(self.raw_recv(recv_chunk), encoding)

        if self.TIMEOUT is None or self.TIMEOUT < 0:
            stop = float('+inf')
        else:
            stop = time.time() + self.TIMEOUT

        received = self._received
        eom = False
        while not (termination in received or eom):
            if time.time() > stop:
                raise LantzTimeoutError
            raw_received = self.raw_recv(recv_chunk)
            received += str(raw_received, encoding)

        self.log_debug('Received {!r} (len={})', received, len(received))

        received, self._received = received.split(termination, 1)

        return received

    def query(self, command, *, send_args=(None, None), recv_args=(None, None)):
        """Send query to the instrument and return the answer

        :param command: command to be sent to the instrument
        :type command: string

        :param send_args: (termination, encoding) to override class defaults
        :param recv_args: (termination, encoding) to override class defaults
        """

        self.send(command, *send_args)
        return self.recv(*recv_args)

    def parse_query(self, command, *,
                    send_args=(None, None), recv_args=(None, None),
                    format=None):
        """Send query to the instrument, parse the output using format
        and return the answer.

        .. seealso:: TextualMixin.query and stringparser
        """
        ans = self.query(command, send_args=send_args, recv_args=recv_args)
        if format:
            parser = self.PARSERS.setdefault(format, ParseProcessor(format))
            ans = parser(ans)
        return ans



"""
    lantz.drivers.keithley.smu2400
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implementation of the 2400 series Source Meter Unit

    Author: Alexandre Bourassa
    Date: 09/08/2016
"""


from lantz import Action, Feat, DictFeat, ureg
from lantz.messagebased import MessageBasedDriver

class SMU2450(MessageBasedDriver):

        DEFAULTS = {
            'COMMON': {
                'write_termination': '\n',
                'read_termination': '\n',
            }
        }

        @Feat(read_once=True)
        def idn(self):
            return self.query('*IDN?')


        @Feat(values={False:'0', True:'1'})
        def output(self):
            return self.query(':OUTP:STAT?')

        @output.setter
        def output(self, value):
            return self.write(':OUTP:STAT {}'.format(value))


        @Feat(units='A', limits=(-1.05, 1.05,))
        def current_compliance(self):
            return self.query(':SOUR:VOLT:ILIM?')

        @current_compliance.setter
        def current_compliance(self, value):
            return self.write(':SOUR:VOLT:ILIM {}'.format(value))


        @Feat(units='V', limits=(-210., 210.,))
        def voltage_compliance(self):
            return self.query(':SOUR:CURR:VLIM?')

        @voltage_compliance.setter
        def voltage_compliance(self, value):
            return self.write(':SOUR:CURR:VLIM {}'.format(value))


        # @Feat(units='V', limits=(-210., 210.,))
        # def source_voltage_range(self):
        #     return self.query(':SOUR:VOLT:RANG?')
        #
        # @source_voltage_range.setter
        # def source_voltage_range(self, value):
        #     return self.write(':SOUR:VOLT:RANG {}'.format(value))
        #
        #
        # @Feat(units='A', limits=(-1.05, 1.05,))
        # def source_current_range(self):
        #     return self.query(':SOUR:CURR:RANG?')
        #
        # @source_current_range.setter
        # def source_current_range(self, value):
        #     return self.write(':SOUR:CURR:RANG {}'.format(value))


        @Feat(units='A', limits=(-1.05, 1.05,))
        def source_current(self):
            return self.query(':SOUR:CURR?')

        @source_current.setter
        def source_current(self, value):
            return self.write(':SOUR:CURR {}'.format(value))


        @Feat(units='V', limits=(-210., 210.,))
        def source_voltage(self):
            return self.query(':SOUR:VOLT?')

        @source_voltage.setter
        def source_voltage(self, value):
            return self.write(':SOUR:VOLT {}'.format(value))


        @Feat(values={'current':'CURR', 'voltage':'VOLT', 'memory':'MEM'})
        def source_function(self):
            return self.query(':SOUR:FUNC?')

        @source_function.setter
        def source_function(self, value):
            return self.write(':SOUR:FUNC {}'.format(value))

        @Action()
        def trigger(self):
            self.write('TRACe:TRIGger "defbuffer1"')

        @Action()
        def reset(self):
            return self.write('*RST')

        @Action()
        def get_data(self):
            """
            Perform a read operation that returns a dictionary
            """
            ans = list(map(float, self.query(':READ?').split(',')))
            return {'voltage': ans[0], 'current': ans[1], 'resistance': ans[2], 'time': ans[3], 'status': ans[4]}


from lantz.foreign import LibraryDriver
from lantz import Feat, DictFeat, Action, Q_

import time
from ctypes import c_uint, c_void_p, c_double, pointer, POINTER

class ANC350(LibraryDriver):

    LIBRARY_NAME = 'anc350v3.dll'
    LIBRARY_PREFIX = 'ANC_'

    RETURN_STATUS = {0:'ANC_Ok', -1:'ANC_Error', 1:"ANC_Timeout", 2:"ANC_NotConnected", 3:"ANC_DriverError",
                     7:"ANC_DeviceLocked", 8:"ANC_Unknown", 9:"ANC_NoDevice", 10:"ANC_NoAxis",
                     11:"ANC_OutOfRange", 12:"ANC_NotAvailable"}

    def __init__(self):
        super(ANC350, self).__init__()

        #Discover systems
        ifaces = c_uint(0x01) # USB interface
        devices = c_uint()
        self.check_error(self.lib.discover(ifaces, pointer(devices)))
        if not devices.value:
            raise RuntimeError('No controller found. Check if controller is connected or if another application is using the connection')
        self.dev_no = c_uint(devices.value - 1)
        self.device = None

        # Function definitions
        self.lib.getFrequency.argtypes = [c_void_p, c_uint, POINTER(c_double)]
        self.lib.setFrequency.argtypes = [c_void_p, c_uint, c_double]
        self.lib.setDcVoltage.argtypes = [c_void_p, c_uint, c_double]
        self.lib.getPosition.argtypes = [c_void_p, c_uint, POINTER(c_double)]
        self.lib.measureCapacitance.argtypes = [c_void_p, c_uint, POINTER(c_double)]
        self.lib.startContinousMove.argtypes = [c_void_p, c_uint, c_uint, c_uint]
        self.lib.setTargetPosition.argtypes = [c_void_p, c_uint, c_double]
        self.lib.setTargetRange.argtypes = [c_void_p, c_uint, c_double]
        self.lib.disconnect.argtypes = [c_void_p]
        return

    def initialize(self, devNo=None):
        if not devNo is None: self.devNo = devNo
        device = c_void_p()
        self.check_error(self.lib.connect(self.dev_no, pointer(device)))
        self.device = device

        #Wait until we get something else then 0 on the position
        while(self.position[2] == Q_(0, 'um')):time.sleep(0.025)

    def finalize(self):
        self.check_error(self.lib.disconnect(self.device))
        self.device = None

    def check_error(self, err):
        if err != 0:
            raise Exception("Driver Error {}: {}".format(err, self.RETURN_STATUS[err]))
        return

    @DictFeat(units='Hz')
    def frequency(self, axis):
        ret_freq = c_double()
        self.check_error(self.lib.getFrequency(self.device, axis, pointer(ret_freq)))
        return ret_freq.value

    @frequency.setter
    def frequency(self, axis, freq):
        self.check_error(self.lib.setFrequency(self.device, axis, freq))
        return err

    @DictFeat(units='um')
    def position(self, axis):
        ret_pos = c_double()
        self.check_error(self.lib.getPosition(self.device, axis, pointer(ret_pos)))
        return ret_pos.value * 1e6

    @position.setter
    def position(self, axis, pos):
        return self.absolute_move(axis, pos*1e-6)


    @DictFeat(units='F')
    def capacitance(self, axis):
        ret_c = c_double()
        self.check_error(self.lib.measureCapacitance(self.device, axis, pointer(ret_c)))
        return ret_c.value

    @DictFeat()
    def status(self, axis):
        status_names = [
            'connected',
            'enabled',
            'moving',
            'target',
            'eot_fwd',
            'eot_bwd',
            'error',
        ]
        status_flags = [c_uint() for _ in range(7)]
        status_flags_p = [pointer(flag) for flag in status_flags]
        self.check_error(self.lib.getAxisStatus(self.device, axis, *status_flags_p))

        ret = dict()
        for status_name, status_flag in zip(status_names, status_flags):
            ret[status_name] = True if status_flag.value else False
        return ret

    # Untested
    @Action()
    def stop(self):
        for axis in range(3):
            self.lib.startContinousMove(self.device, axis, 0, 1)


    @Action()
    def jog(self, axis, speed):
        backward = 0 if speed >= 0.0 else 1
        start = 1 if speed != 0.0 else 0
        self.check_error(self.lib.startContinousMove(self.device, axis, start, backward))
        return

    @Action()
    def single_step(self, axis, direction):
        backward = direction <= 0
        self.lib.startSingleStep(self.device, axis, backward)
        return

    MAX_ABSOLUTE_MOVE = Q_(40, 'um')
    @Action()
    def absolute_move(self, axis, target, max_move=MAX_ABSOLUTE_MOVE):
        if not max_move is None:
            if abs(self.position[axis]-Q_(target, 'm')) > max_move:
                raise Exception("Relative move (target-current) is greater then the max_move")
        self.check_error(self.lib.setTargetPosition(self.device, axis, target))
        enable = 0x01
        relative = 0x00
        self.check_error(self.lib.startAutoMove(self.device, axis, enable, relative))
        return

    MAX_RELATIVE_MOVE = Q_(10e-6, 'um')
    @Action()
    def relative_move(self, axis, delta):
        delta = Q_(delta, 'um')
        if abs(delta) > MAX_RELATIVE_MOVE:
            raise Exception("Relative move <delta> is greater then the MAX_RELATIVE_MOVE")
        else:
            target = self.position + delta
            target = target.to('m').magnitude
            print(target)
            # self.absolute_move(axis, target)


    @Action()
    def relative_move(self, axis, delta, max_move=MAX_RELATIVE_MOVE):
        target = self.position[axis] + delta
        target = target.to('m').magnitude
        print("Moving to {}".format(target))
        self.absolute_move(axis, target, max_move=max_move)

    @Action()
    def set_target_range(self, axis, target_range):
        self.check_error(self.lib.setTargetRange(self.device, axis, target_range))
        return

    @Action()
    def dc_bias(self, axis, voltage):
        self.check_error(self.lib.setDcVoltage(self.device, axis, voltage))
        return

    # ----------------------------------------------
    # Closed-loop Actions
    # These action are much slower but they ensure the move completed
    @Action(units=(None, 'um', 'um', None, 'seconds', None, None))
    def cl_move(self, axis, pos, delta_z=Q_(0.1,'um'), iter_n=10, delay=Q_(0.01, 's'), debug=False, max_iter=1000):
        i = 0
        while(not self.at_pos(Q_(pos, 'um'), delta_z=Q_(delta_z, 'um'), iter_n=iter_n, delay=Q_(delay,'s'))):
            self.position[axis] = Q_(pos, 'um')
            i += 1
            if i>=max_iter:
                raise Exception("Reached max_iter")
        if debug: print("It took {} iterations to move to position".format(i))
        return

    @Action(units=(None, 'um', 'um', None, 'seconds'))
    def at_pos(self, axis, pos, delta_z=Q_(0.1,'um'), iter_n=10, delay=Q_(0.01, 's')):
        for i in range(iter_n):
            time.sleep(delay)
            if abs(self.position[axis].to('um').magnitude-pos)>delta_z:
                return False
        return True
    # ----------------------------------------------

    # Untested
    @Action()
    def stop(self):
        for axis in range(3):
            self.check_error(self.lib.startContinousMove(self.device, axis, 0, 1))
